# This is the Python implementation of the peakdet algorithm.
#
import warnings
import numpy as np


def find_peak_prominence(arr, max_dist=None):
    """Find the local maxima of an array and their prominence.
    The prominence of a peak is defined as the maximal difference between the
    height of the peak and the lowest point in the range until a higher peak.

    Parameters
    ----------
    arr : array_like
    max_dist : int
        If specified, the distance to the adjacent higher peaks is limited
        by `max_dist`.

    Returns
    -------
    loc_max_poss : numpy.array
        The positions of local maxima of a given array.

    proms : numpy.array
        The prominence of the detected maxima.
    """

    arr = np.asarray(arr)
    n = len(arr)
    max_dist = len(arr) if max_dist is None else int(max_dist)

    # Finding all local minima and maxima (i.e. points the are lower/higher than
    # both immediate non-nan neighbors).
    arr_nonans = arr[~np.isnan(arr)]
    idxs_nonans2idx = np.arange(arr.size)[~np.isnan(arr)]

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)

        is_min_left = np.r_[False, arr_nonans[:-1] > arr_nonans[1:]]
        is_min_right = np.r_[arr_nonans[:-1] < arr_nonans[1:], False]
        is_loc_min = is_min_left & is_min_right
        loc_min_poss = np.where(is_loc_min)[0]
        loc_min_poss = idxs_nonans2idx[loc_min_poss]

        is_max_left = np.r_[False, arr_nonans[:-1] < arr_nonans[1:]]
        is_max_right = np.r_[arr_nonans[:-1] > arr_nonans[1:], False]
        is_loc_max = is_max_left & is_max_right
        loc_max_poss = np.where(is_loc_max)[0]
        loc_max_poss = idxs_nonans2idx[loc_max_poss]

    # For each maximum, find the position of a higher peak on the left and
    # on the right. If there are no higher peaks within the `max_dist` range,
    # just use the position `max_dist` away.
    left_maxs = -1 * np.ones(len(loc_max_poss), dtype=np.int)
    right_maxs = -1 * np.ones(len(loc_max_poss), dtype=np.int)

    for i, pos in enumerate(loc_max_poss):
        for j in range(pos - 1, -1, -1):
            if (arr[j] > arr[pos]) or (pos - j > max_dist):
                left_maxs[i] = j
                break

        for j in range(pos + 1, n):
            if (arr[j] > arr[pos]) or (j - pos > max_dist):
                right_maxs[i] = j
                break

    # Find the prominence of each peak with respect of the lowest point
    # between the peak and the adjacent higher peaks, on the left and the right
    # separately.
    left_max_proms = np.array(
        [
            (
                arr[pos] - np.nanmin(arr[left_maxs[i] : pos])
                if (left_maxs[i] >= 0)
                else np.nan
            )
            for i, pos in enumerate(loc_max_poss)
        ]
    )

    right_max_proms = np.array(
        [
            (
                arr[pos] - np.nanmin(arr[pos : right_maxs[i]])
                if (right_maxs[i] >= 0)
                else np.nan
            )
            for i, pos in enumerate(loc_max_poss)
        ]
    )

    # In 1D, the topographic definition of the prominence of a peak reduces to
    # the minimum of the left-side and right-side prominence.

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", RuntimeWarning)

        max_proms = np.nanmin(np.vstack([left_max_proms, right_max_proms]), axis=0)

    # The global maximum, by definition, does not have higher peaks around it and
    # thus its prominence is explicitly defined with respect to the lowest local
    # minimum. This issue arises only if max_dist was not specified, otherwise
    # the prominence of the global maximum is already calculated with respect
    # to the lowest point within the `max_dist` range.
    # If no local minima are within the `max_dist` range, just use the
    # lowest point.
    global_max_mask = (left_maxs == -1) & (right_maxs == -1)
    if (global_max_mask).sum() > 0:
        global_max_idx = np.where(global_max_mask)[0][0]
        global_max_pos = loc_max_poss[global_max_idx]
        neighbor_loc_mins = (loc_min_poss >= global_max_pos - max_dist) & (
            loc_min_poss < global_max_pos + max_dist
        )
        if np.any(neighbor_loc_mins):
            max_proms[global_max_idx] = arr[global_max_pos] - np.nanmin(
                arr[loc_min_poss[neighbor_loc_mins]]
            )
        else:
            max_proms[global_max_idx] = arr[global_max_pos] - np.nanmin(
                arr[max(global_max_pos - max_dist, 0) : global_max_pos + max_dist]
            )

    return loc_max_poss, max_proms


def peakdet(arr, min_prominence):
    """Detect local peaks in an array.
    Finds a sequence of minima and maxima such that the two consecutive extrema
    have a value difference (i.e. a prominence) >= `min_prominence`. This is
    analogous to the definition of prominence in topography:
    https://en.wikipedia.org/wiki/Topographic_prominence

    The original peakdet algorithm was designed by Eli Billauer and described in
    http://billauer.co.il/peakdet.html (v. 3.4.05, Explicitly not copyrighted).
    This function is released to the public domain; Any use is allowed.
    The Python implementation was published
    by endolith on Github: https://gist.github.com/endolith/250860 .

    Here, we use the endolith's implementation with minimal to none modifications
    to the algorithm, but with significant changes in the interface and
    the documentation

    Parameters
    ----------
    arr : array_like
    min_prominence : float
        The minimal prominence of detected extrema.

    Returns
    -------
    maxidxs, minidx : numpy.array
        The indices of the maxima and minima in `arr`.

    """
    maxidxs = []
    minidxs = []

    x = np.arange(len(arr))

    arr = np.asarray(arr)

    if not np.isscalar(min_prominence):
        raise Exception("Input argument delta must be a scalar")

    if min_prominence <= 0:
        raise Exception("Input argument delta must be positive")

    mn, mx = np.inf, -np.inf
    mnpos, mxpos = np.nan, np.nan

    lookformax = True

    for i in np.arange(len(arr)):
        this = arr[i]
        if this > mx:
            mx = this
            mxpos = x[i]
        if this < mn:
            mn = this
            mnpos = x[i]

        if lookformax:
            if this < mx - min_prominence:
                maxidxs.append(mxpos)
                mn = this
                mnpos = x[i]
                lookformax = False
        else:
            if this > mn + min_prominence:
                minidxs.append(mnpos)
                mx = this
                mxpos = x[i]
                lookformax = True

    return np.array(minidxs), np.array(maxidxs)


def find_peak_prominence_iterative(
    arr,
    min_prom=None,
    max_prom=None,
    steps_prom=1000,
    log_space_proms=True,
    min_n_peak_pairs=5,
):
    """Finds the minima/maxima of an array using the peakdet algorithm at
    different values of the threshold prominence. For each location, returns
    the maximal threshold prominence at which it is called as a minimum/maximum.

    Note that this function is inferior in every aspect to find_peak_prominence.
    We keep it for testing purposes and will remove in the future.

    Parameters
    ----------
    arr : array_like
    min_prom, max_prom : float
        The minimal and the maximal values of prominence to probe.
        If None, these values are inferred as the minimal and the maximal
        non-zero difference between any two elements of `v`.
    steps_prom : int
        The number of threshold prominence values to probe in the range
        between `min_prom` and `max_prom`.
    log_space_proms : bool
        If True, probe logarithmically spaced values of the threshold prominence
        in the range between `min_prom` and `max_prom`.
    min_n_peak_pairs : int
        If the number of detected minima/maxima at a certain threshold
        prominence is < `min_n_peak_pairs`, the detected peaks are ignored.

    Returns
    -------
    minproms, maxproms : numpy.array
        The prominence of detected minima and maxima.
    """
    if ((min_prom is None) and (max_prom is not None)) or (
        (min_prom is not None) and (max_prom is None)
    ):
        raise Exception(
            "Please, provide either both min_prom and max_prom or "
            "none to infer these values from the data."
        )

    if (min_prom is None) and (max_prom is None):
        arr_sorted = np.sort(arr)
        arr_sorted = arr_sorted[np.isfinite(arr_sorted)]
        max_prom = arr_sorted[-1] - arr_sorted[0]

        diffs = np.diff(arr_sorted)
        min_prom = diffs[diffs != 0].min()

    if log_space_proms:
        proms = 10 ** np.linspace(np.log10(min_prom), np.log10(max_prom), steps_prom)
    else:
        proms = np.linspace(min_prom, max_prom, steps_prom)

    minproms = np.nan * np.ones_like(arr)
    maxproms = np.nan * np.ones_like(arr)
    for p in proms:
        minidxs, maxidxs = peakdet(arr, p)
        if (len(minidxs) >= min_n_peak_pairs) and (len(minidxs) >= min_n_peak_pairs):

            valid_mins = minidxs[np.isfinite(arr[minidxs])]
            minproms[valid_mins] = p

            valid_maxs = maxidxs[np.isfinite(arr[maxidxs])]
            maxproms[valid_maxs] = p

    return minproms, maxproms



from scipy.linalg import toeplitz
import numpy as np
from cooltools.lib.numutils import LazyToeplitz


n = 100
m = 150
c = np.arange(1, n + 1)
r = np.r_[1, np.arange(-2, -m, -1)]

L = LazyToeplitz(c, r)
T = toeplitz(c, r)


def test_symmetric():
    for si in [
        slice(10, 20),
        slice(0, 150),
        slice(0, 0),
        slice(150, 150),
        slice(10, 10),
    ]:
        assert np.allclose(L[si, si], T[si, si])


def test_triu_no_overlap():
    for si, sj in [
        (slice(10, 20), slice(30, 40)),
        (slice(10, 15), slice(30, 40)),
        (slice(10, 20), slice(30, 45)),
    ]:
        assert np.allclose(L[si, sj], T[si, sj])


def test_tril_no_overlap():
    for si, sj in [
        (slice(30, 40), slice(10, 20)),
        (slice(30, 40), slice(10, 15)),
        (slice(30, 45), slice(10, 20)),
    ]:
        assert np.allclose(L[si, sj], T[si, sj])


def test_triu_with_overlap():
    for si, sj in [
        (slice(10, 20), slice(15, 25)),
        (slice(13, 22), slice(15, 25)),
        (slice(10, 20), slice(18, 22)),
    ]:
        assert np.allclose(L[si, sj], T[si, sj])


def test_tril_with_overlap():
    for si, sj in [
        (slice(15, 25), slice(10, 20)),
        (slice(15, 22), slice(10, 20)),
        (slice(15, 25), slice(10, 18)),
    ]:
        assert np.allclose(L[si, sj], T[si, sj])


def test_nested():
    for si, sj in [
        (slice(10, 40), slice(20, 30)),
        (slice(10, 35), slice(20, 30)),
        (slice(10, 40), slice(20, 25)),
        (slice(20, 30), slice(10, 40)),
    ]:
        assert np.allclose(L[si, sj], T[si, sj])



# -*- coding: utf-8 -*-
from __future__ import division, print_function
import click
import sys
from .. import __version__


# Monkey patch
click.core._verify_python3_env = lambda: None


CONTEXT_SETTINGS = {
    "help_option_names": ["-h", "--help"],
}


@click.version_option(__version__, "-V", "--version")
@click.group(context_settings=CONTEXT_SETTINGS)
@click.option(
    "-v", "--verbose",
    help="Verbose logging",
    is_flag=True,
    default=False
)
@click.option(
    "-d", "--debug",
    help="Post mortem debugging",
    is_flag=True,
    default=False
)
def cli(verbose, debug):
    """
    Type -h or --help after any subcommand for more information.

    """
    if verbose:
        pass
        # logger.setLevel(logging.DEBUG)

    if debug:
        import traceback

        try:
            import ipdb as pdb
        except ImportError:
            import pdb

        def _excepthook(exc_type, value, tb):
            traceback.print_exception(exc_type, value, tb)
            print()
            pdb.pm()

        sys.excepthook = _excepthook


from . import (
    dump_cworld,
    diamond_insulation,
    compute_expected,
    logbin_expected,
    compute_saddle,
    call_dots,
    call_compartments,
    genome,
    random_sample,
)



import cooler
import bioframe
import os.path as op

import cooltools.snipping
import numpy as np
import pandas as pd


def test_ondiag_pileups_with_expected(request):
    """
    Test the snipping on matrix:
    """
    # Read cool file and create regions out of it:
    clr = cooler.Cooler(op.join(request.fspath.dirname, "data/CN.mm9.1000kb.cool"))
    exp = pd.read_table(op.join(request.fspath.dirname, "data/CN.mm9.toy_expected.tsv"))
    regions = bioframe.read_table(
        op.join(request.fspath.dirname, "data/CN.mm9.toy_regions.bed"), schema="bed4"
    )
    for snipper_class in (
        cooltools.snipping.ObsExpSnipper,
        cooltools.snipping.ExpectedSnipper,
    ):
        snipper = snipper_class(clr, exp, regions=regions)

        # I.
        # Example region with windows, two regions from annotated genomic regions:
        windows = cooltools.snipping.make_bin_aligned_windows(
            1_000_000, ["chr1", "chr1"], [102_000_000, 105_000_000], flank_bp=2_000_000
        )
        windows = cooltools.snipping.assign_regions(windows, regions).reset_index(
            drop=True
        )
        stack = cooltools.snipping.pileup(
            windows, snipper.select, snipper.snip, map=map
        )

        # Check that the size of snips is OK and there are two of them:
        assert stack.shape == (5, 5, 2)

        # II.
        # Example region with windows, second window comes from unannotated genomic region:
        windows = cooltools.snipping.make_bin_aligned_windows(
            1_000_000, ["chr1", "chr1"], [120_000_000, 160_000_000], flank_bp=2_000_000
        )
        windows = cooltools.snipping.assign_regions(windows, regions).reset_index(
            drop=True
        )

        stack = cooltools.snipping.pileup(
            windows, snipper.select, snipper.snip, map=map
        )

        assert stack.shape == (5, 5, 2)
        assert np.all(np.isnan(stack[:, :, 1]))


def test_ondiag_pileups_without_expected(request):
    """
    Test the snipping on matrix:
    """
    # Read cool file and create regions out of it:
    clr = cooler.Cooler(op.join(request.fspath.dirname, "data/CN.mm9.1000kb.cool"))
    regions = bioframe.read_table(
        op.join(request.fspath.dirname, "data/CN.mm9.toy_regions.bed"), schema="bed4"
    )

    # I.
    # Example region with windows, two regions from annotated genomic regions:
    windows = cooltools.snipping.make_bin_aligned_windows(
        1_000_000, ["chr1", "chr1"], [120_000_000, 160_000_000], flank_bp=2_000_000
    )

    windows = cooltools.snipping.assign_regions(windows, regions).reset_index(drop=True)

    snipper = cooltools.snipping.CoolerSnipper(clr, regions=regions)
    stack = cooltools.snipping.pileup(windows, snipper.select, snipper.snip, map=map)

    # Check that the size of snips is OK and there are two of them:
    assert stack.shape == (5, 5, 2)

    # II.
    # Example region with windows, second window comes from unannotated genomic region:
    windows = cooltools.snipping.make_bin_aligned_windows(
        1_000_000, ["chr1", "chr1"], [120_000_000, 160_000_000], flank_bp=2_000_000
    )
    windows = cooltools.snipping.assign_regions(windows, regions).reset_index(drop=True)

    stack = cooltools.snipping.pileup(windows, snipper.select, snipper.snip, map=map)

    assert stack.shape == (5, 5, 2)
    assert np.all(np.isfinite(stack[:, :, 0]))
    assert np.all(np.isnan(stack[:, :, 1]))


def test_offdiag_pileups_with_expected(request):
    """
    Test the snipping on matrix:
    """
    # Read cool file and create regions out of it:
    clr = cooler.Cooler(op.join(request.fspath.dirname, "data/CN.mm9.1000kb.cool"))
    exp = pd.read_table(op.join(request.fspath.dirname, "data/CN.mm9.toy_expected.tsv"))
    regions = bioframe.read_table(
        op.join(request.fspath.dirname, "data/CN.mm9.toy_regions.bed"), schema="bed4"
    )
    for snipper_class in (
        cooltools.snipping.ObsExpSnipper,
        cooltools.snipping.ExpectedSnipper,
    ):

        snipper = snipper_class(clr, exp, regions=regions)

        # I.
        # Example region with windows, two off-diagonal features from annotated genomic regions:
        windows1 = cooltools.snipping.make_bin_aligned_windows(
            1_000_000, ["chr1", "chr1"], [102_000_000, 105_000_000], flank_bp=2_000_000
        )
        windows2 = cooltools.snipping.make_bin_aligned_windows(
            1_000_000, ["chr1", "chr1"], [105_000_000, 109_000_000], flank_bp=2_000_000
        )
        windows = pd.merge(
            windows1, windows2, left_index=True, right_index=True, suffixes=("1", "2")
        )
        windows = cooltools.snipping.assign_regions(windows, regions).reset_index(
            drop=True
        )

        stack = cooltools.snipping.pileup(
            windows, snipper.select, snipper.snip, map=map
        )

        # Check that the size of snips is OK and there are two of them:
        assert stack.shape == (5, 5, 2)

        # II.
        # Example region with windows, second window is between two different regions:
        windows1 = cooltools.snipping.make_bin_aligned_windows(
            1_000_000, ["chr1", "chr1"], [102_000_000, 10_000_000], flank_bp=2_000_000
        )
        windows2 = cooltools.snipping.make_bin_aligned_windows(
            1_000_000, ["chr1", "chr1"], [105_000_000, 109_000_000], flank_bp=2_000_000
        )
        windows = pd.merge(
            windows1, windows2, left_index=True, right_index=True, suffixes=("1", "2")
        )
        windows = cooltools.snipping.assign_regions(windows, regions).reset_index(
            drop=True
        )

        stack = cooltools.snipping.pileup(
            windows, snipper.select, snipper.snip, map=map
        )

        assert stack.shape == (5, 5, 2)
        assert np.all(np.isnan(stack[:, :, 1]))


def test_offdiag_pileups_without_expected(request):
    """
    Test the snipping on matrix:
    """
    # Read cool file and create regions out of it:
    clr = cooler.Cooler(op.join(request.fspath.dirname, "data/CN.mm9.1000kb.cool"))
    regions = bioframe.read_table(
        op.join(request.fspath.dirname, "data/CN.mm9.toy_regions.bed"), schema="bed4"
    )

    # I.
    # Example region with windows, two regions from annotated genomic regions:
    windows1 = cooltools.snipping.make_bin_aligned_windows(
        1_000_000, ["chr1", "chr1"], [102_000_000, 105_000_000], flank_bp=2_000_000
    )
    windows2 = cooltools.snipping.make_bin_aligned_windows(
        1_000_000, ["chr1", "chr1"], [105_000_000, 109_000_000], flank_bp=2_000_000
    )
    windows = pd.merge(
        windows1, windows2, left_index=True, right_index=True, suffixes=("1", "2")
    )
    windows = cooltools.snipping.assign_regions(windows, regions).reset_index(drop=True)

    snipper = cooltools.snipping.CoolerSnipper(clr, regions=regions)
    stack = cooltools.snipping.pileup(windows, snipper.select, snipper.snip, map=map)

    # Check that the size of snips is OK and there are two of them:
    assert stack.shape == (5, 5, 2)

    # II.
    # Example region with windows, second window comes from unannotated genomic region:
    windows1 = cooltools.snipping.make_bin_aligned_windows(
        1_000_000, ["chr1", "chr1"], [102_000_000, 10_000_000], flank_bp=2_000_000
    )
    windows2 = cooltools.snipping.make_bin_aligned_windows(
        1_000_000, ["chr1", "chr1"], [105_000_000, 109_000_000], flank_bp=2_000_000
    )
    windows = pd.merge(
        windows1, windows2, left_index=True, right_index=True, suffixes=("1", "2")
    )
    windows = cooltools.snipping.assign_regions(windows, regions).reset_index(drop=True)

    stack = cooltools.snipping.pileup(windows, snipper.select, snipper.snip, map=map)

    assert stack.shape == (5, 5, 2)
    assert np.all(np.isfinite(stack[:, :, 0]))
    assert np.all(np.isnan(stack[:, :, 1]))


def test_snipper_with_regions_and_expected(request):
    clr = cooler.Cooler(op.join(request.fspath.dirname, "data/CN.mm9.1000kb.cool"))
    exp = pd.read_table(op.join(request.fspath.dirname, "data/CN.mm9.toy_expected.tsv"))
    regions = bioframe.read_table(
        op.join(request.fspath.dirname, "data/CN.mm9.toy_regions.bed"), schema="bed4"
    )
    for snipper_class in (
        cooltools.snipping.ObsExpSnipper,
        cooltools.snipping.ExpectedSnipper,
    ):
        snipper = snipper_class(clr, exp, regions=regions)
        matrix = snipper.select("foo", "foo")
        snippet = snipper.snip(
            matrix, "foo", "foo", (110_000_000, 120_000_000, 110_000_000, 120_000_000)
        )
        assert snippet.shape is not None


def test_cooler_snipper_with_regions(request):
    clr = cooler.Cooler(op.join(request.fspath.dirname, "data/CN.mm9.1000kb.cool"))
    regions = bioframe.read_table(
        op.join(request.fspath.dirname, "data/CN.mm9.toy_regions.bed"), schema="bed4"
    )
    snipper = cooltools.snipping.CoolerSnipper(clr, regions=regions)
    matrix = snipper.select("foo", "foo")
    snippet = snipper.snip(
        matrix, "foo", "foo", (110_000_000, 120_000_000, 110_000_000, 120_000_000)
    )
    assert snippet.shape is not None



from itertools import combinations
from functools import partial
from scipy.linalg import toeplitz
from cytoolz import merge
import numpy as np
import pandas as pd
from .lib import numutils

import bioframe


def ecdf(x, v, side="left"):
    """
    Return array `x`'s empirical CDF value(s) at the points in `v`.
    This is based on the :func:`statsmodels.distributions.ECDF` step function.
    This is the inverse of `quantile`.

    """
    x = np.asarray(x)
    ind = np.searchsorted(np.sort(x), v, side=side) - 1
    y = np.linspace(1.0 / len(x), 1.0, len(x))
    return y[ind]


def quantile(x, q, **kwargs):
    """
    Return the values of the quantile cut points specified by fractions `q` of
    a sequence of data given by `x`.

    """
    x = np.asarray(x)
    p = np.asarray(q) * 100
    return np.nanpercentile(x, p, **kwargs)


def mask_bad_bins(track, bintable):
    """
    Mask (set to NaN) values in track where bin is masked in bintable.

    Parameters
    ----------
    track : tuple of (DataFrame, str)
        bedGraph-like dataframe along with the name of the value column.
    bintable : tuple of (DataFrame, str)
        bedGraph-like dataframe along with the name of the weight column.

    Returns
    -------
    track : DataFrame
        New bedGraph-like dataframe with bad bins masked in the value column
    """
    track, name = track

    bintable, weight_name = bintable

    track = pd.merge(
        track[["chrom", "start", "end", name]], bintable, on=["chrom", "start", "end"]
    )
    track.loc[~np.isfinite(track[weight_name]), name] = np.nan
    track = track[["chrom", "start", "end", name]]

    return track


def digitize_track(binedges, track, regions=None):
    """
    Digitize genomic signal tracks into integers between `1` and `n`.

    Parameters
    ----------
    binedges : 1D array (length n + 1)
        Bin edges for quantization of signal. For `n` bins, there are `n + 1`
        edges. See encoding details in Notes.
    track : tuple of (DataFrame, str)
        bedGraph-like dataframe along with the name of the value column.
    regions: sequence of str or tuples
        List of genomic regions to include. Each can be a chromosome, a
        UCSC-style genomic region string or a tuple.

    Returns
    -------
    digitized : DataFrame
        New bedGraph-like dataframe with value column and an additional
        digitized value column with name suffixed by '.d'
    hist : 1D array (length n + 2)
        Histogram of digitized signal values. Its length is `n + 2` because
        the first and last elements correspond to outliers. See notes.

    Notes
    -----
    The digital encoding is as follows:

    - `1..n` <-> values assigned to histogram bins
    - `0` <-> left outlier values
    - `n+1` <-> right outlier values
    - `-1` <-> missing data (NaNs)

    """
    if not isinstance(track, tuple):
        raise ValueError("``track`` should be a tuple of (dataframe, column_name)")
    track, name = track

    # subset and re-order chromosome groups
    if regions is not None:
        regions = [bioframe.parse_region(reg) for reg in regions]
        track = pd.concat(
            bioframe.select(track, region) for region in regions
        )

    # histogram the signal
    digitized = track.copy()
    digitized[name + ".d"] = np.digitize(track[name].values, binedges, right=False)
    mask = track[name].isnull()
    digitized.loc[mask, name + ".d"] = -1
    x = digitized[name + ".d"].values.copy()
    x = x[(x > 0) & (x < len(binedges) + 1)]
    hist = np.bincount(x, minlength=len(binedges) + 1)
    return digitized, hist


def make_cis_obsexp_fetcher(clr, expected, weight_name="weight"):
    """
    Construct a function that returns intra-chromosomal OBS/EXP for symmetrical regions.

    Parameters
    ----------
    clr : cooler.Cooler
        Observed matrix.
    expected : tuple of (DataFrame, str)
        Diagonal summary statistics for each chromosome, and name of the column
        with the values of expected to use.
    weight_name : str
        Name of the column in the clr.bins to use as balancing weights

    Returns
    -------
    getexpected(reg, _). 2nd arg is ignored.

    """
    expected, expected_name = expected
    expected = {k: x.values for k, x in expected.groupby("region")[expected_name]}

    def _fetch_cis_oe(reg1, reg2):
        obs_mat = clr.matrix(balance=weight_name).fetch(reg1)
        exp_mat = toeplitz(expected[reg1][: obs_mat.shape[0]])
        return obs_mat / exp_mat

    return _fetch_cis_oe


def make_trans_obsexp_fetcher(clr, expected, weight_name="weight"):
    """
    Construct a function that returns OBS/EXP for any pair of chromosomes.

    Parameters
    ----------
    clr : cooler.Cooler
        Observed matrix.
    expected : (DataFrame, name) or scalar
        Average trans values. If a scalar, it is assumed to be a global trans
        expected value. If a tuple of (dataframe, name), the dataframe must
        have a MultiIndex with 'region1' and 'region2' and must also have a column
        labeled ``name``, with the values of expected.
    weight_name : str
        Name of the column in the clr.bins to use as balancing weights

    Returns
    -----
    getexpected(reg1, reg2)

    """

    if np.isscalar(expected):
        return lambda reg1, reg2: (
            clr.matrix(balance=weight_name).fetch(reg1, reg2) / expected
        )

    elif isinstance(expected, (tuple, list)):
        expected, expected_name = expected

        if not expected_name:
            raise ValueError("Name of data column not provided.")

        expected = {
            k: x.values for k, x in expected.groupby(["region1", "region2"])[expected_name]
        }

        def _fetch_trans_exp(region1, region2):
            # Handle region flipping
            if (region1, region2) in expected.keys():
                return expected[region1, region2]
            elif (region2, region1) in expected.keys():
                return expected[region2, region1]
            # .loc is the right way to get [region1,region2] value from MultiIndex df:
            # https://pandas.pydata.org/pandas-docs/stable/advanced.html#advanced-indexing-with-hierarchical-index
            else:
                raise KeyError(
                    "trans-exp index is missing a pair of chromosomes: "
                    "{}, {}".format(region1, region2)
                )

        def _fetch_trans_oe(reg1, reg2):
            reg1 = bioframe.parse_region(reg1)
            reg2 = bioframe.parse_region(reg2)

            return clr.matrix(balance=weight_name).fetch(reg1, reg2) / _fetch_trans_exp(
                reg1[0], reg2[0]
            )

        return _fetch_trans_oe

    else:
        raise ValueError("Unknown type of expected")


def _accumulate(S, C, getmatrix, digitized, reg1, reg2, min_diag, max_diag, verbose):
    n_bins = S.shape[0]
    matrix = getmatrix(reg1, reg2)

    if reg1[0] == reg2[0]:
        for d in np.arange(-min_diag + 1, min_diag):
            numutils.set_diag(matrix, np.nan, d)
        if max_diag >= 0:
            for d in np.append(
                np.arange(-matrix.shape[0], -max_diag),
                np.arange(max_diag + 1, matrix.shape[0]),
            ):
                numutils.set_diag(matrix, np.nan, d)

    if verbose:
        print("regions {} vs {}".format(reg1, reg2))

    for i in range(n_bins):
        row_mask = digitized[reg1] == i
        for j in range(n_bins):
            col_mask = digitized[reg2] == j
            data = matrix[row_mask, :][:, col_mask]
            data = data[np.isfinite(data)]
            S[i, j] += np.sum(data)
            C[i, j] += float(len(data))


def make_saddle(
    getmatrix,
    binedges,
    digitized,
    contact_type,
    regions=None,
    min_diag=3,
    max_diag=-1,
    trim_outliers=False,
    verbose=False,
):
    """
    Make a matrix of average interaction probabilities between genomic bin
    pairs as a function of a specified genomic track. The provided genomic
    track must be pre-quantized as integers (i.e. digitized).

    Parameters
    ----------
    getmatrix : function
        A function returning a matrix of interaction between two chromosomes
        given their names/indicies.
    binedges : 1D array (length n + 1)
        Bin edges of the digitized signal. For `n` bins, there are `n + 1`
        edges. See :func:`digitize_track`.
    digitized : tuple of (DataFrame, str)
        BedGraph-like dataframe of digitized signal along with the name of
        the digitized value column.
    contact_type : str
        If 'cis' then only cis interactions are used to build the matrix.
        If 'trans', only trans interactions are used.
    regions : sequence of str or tuple, optional
        A list of genomic regions to use. Each can be a chromosome, a
        UCSC-style genomic region string or a tuple.
    min_diag : int
        Smallest diagonal to include in computation. Ignored with
        contact_type=trans.
    max_diag : int
        Biggest diagonal to include in computation. Ignored with
        contact_type=trans.
    trim_outliers : bool, optional
        Remove first and last row and column from the output matrix.
    verbose : bool, optional
        If True then reports progress.

    Returns
    -------
    interaction_sum : 2D array
        The matrix of summed interaction probability between two genomic bins
        given their values of the provided genomic track.
    interaction_count : 2D array
        The matrix of the number of genomic bin pairs that contributed to the
        corresponding pixel of ``interaction_sum``.

    """
    digitized_df, name = digitized
    digitized_df = digitized_df[["chrom", "start", "end", name]]

    if regions is None:
        regions = [
            (chrom, df.start.min(), df.end.max())
            for chrom, df in digitized_df.groupby("chrom")
        ]

    regions = bioframe.parse_regions(regions)

    digitized_tracks = {}
    for reg in regions.values:
        track = bioframe.select(digitized_df, reg)
        digitized_tracks[reg[3]] = track[name]  # 3 = name

    if contact_type == "cis":
        supports = list(zip(regions["name"], regions["name"]))
    elif contact_type == "trans":
        supports = list(combinations(regions["name"], 2))
    else:
        raise ValueError(
            "The allowed values for the contact_type " "argument are 'cis' or 'trans'."
        )

    # n_bins here includes 2 open bins
    # for values <lo and >hi.
    n_bins = len(binedges) + 1
    interaction_sum = np.zeros((n_bins, n_bins))
    interaction_count = np.zeros((n_bins, n_bins))

    for reg1, reg2 in supports:
        _accumulate(
            interaction_sum,
            interaction_count,
            getmatrix,
            digitized_tracks,
            reg1,
            reg2,
            min_diag,
            max_diag,
            verbose,
        )

    interaction_sum += interaction_sum.T
    interaction_count += interaction_count.T

    if trim_outliers:
        interaction_sum = interaction_sum[1:-1, 1:-1]
        interaction_count = interaction_count[1:-1, 1:-1]

    return interaction_sum, interaction_count


def saddleplot(
    binedges,
    counts,
    saddledata,
    cmap="coolwarm",
    scale="log",
    vmin=0.5,
    vmax=2,
    color=None,
    title=None,
    xlabel=None,
    ylabel=None,
    clabel=None,
    fig=None,
    fig_kws=None,
    heatmap_kws=None,
    margin_kws=None,
    cbar_kws=None,
    subplot_spec=None,
):
    """
    Generate a saddle plot.

    Parameters
    ----------
    binedges : 1D array-like
        For `n` bins, there should be `n + 1` bin edges
    counts : 1D array-like
        Signal track histogram produced by `digitize_track`. It will include
        2 flanking elements for outlier values, thus the length should be
        `n + 2`.
    saddledata : 2D array-like
        Saddle matrix produced by `make_saddle`. It will include 2 flanking
        rows/columns for outlier signal values, thus the shape should be
        `(n+2, n+2)`.
    cmap : str or matplotlib colormap
        Colormap to use for plotting the saddle heatmap
    scale : str
        Color scaling to use for plotting the saddle heatmap: log or linear
    vmin, vmax : float
        Value limits for coloring the saddle heatmap
    color : matplotlib color value
        Face color for margin bar plots
    fig : matplotlib Figure, optional
        Specified figure to plot on. A new figure is created if none is
        provided.
    fig_kws : dict, optional
        Passed on to `plt.Figure()`
    heatmap_kws : dict, optional
        Passed on to `ax.imshow()`
    margin_kws : dict, optional
        Passed on to `ax.bar()` and `ax.barh()`
    cbar_kws : dict, optional
        Passed on to `plt.colorbar()`
    subplot_spec : GridSpec object
        Specify a subregion of a figure to using a GridSpec.

    Returns
    -------
    Dictionary of axes objects.

    """
    from matplotlib.gridspec import GridSpec, GridSpecFromSubplotSpec
    from matplotlib.colors import Normalize, LogNorm
    from matplotlib import ticker
    import matplotlib.pyplot as plt

    class MinOneMaxFormatter(ticker.LogFormatter):
        def set_locs(self, locs=None):
            self._sublabels = set([vmin % 10 * 10, vmax % 10, 1])

        def __call__(self, x, pos=None):
            if x not in [vmin, 1, vmax]:
                return ""
            else:
                return "{x:g}".format(x=x)

    n_edges = len(binedges)
    n_bins = n_edges - 1
    lo, hi = binedges[0], binedges[-1]

    # Histogram and saddledata are flanked by outlier bins
    n = saddledata.shape[0]
    X, Y = np.meshgrid(binedges, binedges)
    C = saddledata
    hist = counts
    if (n - n_bins) == 2:
        C = C[1:-1, 1:-1]
        hist = hist[1:-1]

    # Layout
    if subplot_spec is not None:
        GridSpec = partial(GridSpecFromSubplotSpec, subplot_spec=subplot_spec)
    grid = {}
    gs = GridSpec(
        nrows=3,
        ncols=3,
        width_ratios=[0.2, 1, 0.1],
        height_ratios=[0.2, 1, 0.1],
        wspace=0.05,
        hspace=0.05,
    )

    # Figure
    if fig is None:
        fig_kws_default = dict(figsize=(5, 5))
        fig_kws = merge(fig_kws_default, fig_kws if fig_kws is not None else {})
        fig = plt.figure(**fig_kws)

    # Heatmap
    if scale == "log":
        norm = LogNorm(vmin=vmin, vmax=vmax)
    elif scale == "linear":
        norm = Normalize(vmin=vmin, vmax=vmax)
    else:
        raise ValueError("Only linear and log color scaling is supported")

    grid["ax_heatmap"] = ax = plt.subplot(gs[4])
    heatmap_kws_default = dict(cmap="coolwarm", rasterized=True)
    heatmap_kws = merge(
        heatmap_kws_default, heatmap_kws if heatmap_kws is not None else {}
    )
    img = ax.pcolormesh(X, Y, C, norm=norm, **heatmap_kws)
    plt.gca().yaxis.set_visible(False)

    # Margins
    margin_kws_default = dict(edgecolor="k", facecolor=color, linewidth=1)
    margin_kws = merge(margin_kws_default, margin_kws if margin_kws is not None else {})
    # left margin hist
    grid["ax_margin_y"] = plt.subplot(gs[3], sharey=grid["ax_heatmap"])
    plt.barh(
        binedges[:-1], height=np.diff(binedges), width=hist, align="edge", **margin_kws
    )
    plt.xlim(plt.xlim()[1], plt.xlim()[0])  # fliplr
    plt.ylim(hi, lo)
    plt.gca().spines["top"].set_visible(False)
    plt.gca().spines["bottom"].set_visible(False)
    plt.gca().spines["left"].set_visible(False)
    plt.gca().xaxis.set_visible(False)
    # top margin hist
    grid["ax_margin_x"] = plt.subplot(gs[1], sharex=grid["ax_heatmap"])
    plt.bar(
        binedges[:-1], width=np.diff(binedges), height=hist, align="edge", **margin_kws
    )
    plt.xlim(lo, hi)
    # plt.ylim(plt.ylim())  # correct
    plt.gca().spines["top"].set_visible(False)
    plt.gca().spines["right"].set_visible(False)
    plt.gca().spines["left"].set_visible(False)
    plt.gca().xaxis.set_visible(False)
    plt.gca().yaxis.set_visible(False)

    # Colorbar
    grid["ax_cbar"] = plt.subplot(gs[5])
    cbar_kws_default = dict(fraction=0.8, label=clabel or "")
    cbar_kws = merge(cbar_kws_default, cbar_kws if cbar_kws is not None else {})
    if scale == "linear" and vmin is not None and vmax is not None:
        grid["cbar"] = cb = plt.colorbar(img, **cbar_kws)
        # cb.set_ticks(np.arange(vmin, vmax + 0.001, 0.5))
        # # do linspace between vmin and vmax of 5 segments and trunc to 1 decimal:
        decimal = 10
        nsegments = 5
        cd_ticks = np.trunc(np.linspace(vmin, vmax, nsegments) * decimal) / decimal
        cb.set_ticks(cd_ticks)
    else:
        grid["cbar"] = cb = plt.colorbar(img, format=MinOneMaxFormatter(), **cbar_kws)
        cb.ax.yaxis.set_minor_formatter(MinOneMaxFormatter())

    # extra settings
    grid["ax_heatmap"].set_xlim(lo, hi)
    grid["ax_heatmap"].set_ylim(hi, lo)
    plt.grid(False)
    plt.axis("off")
    if title is not None:
        grid["ax_margin_x"].set_title(title)
    if xlabel is not None:
        grid["ax_heatmap"].set_xlabel(xlabel)
    if ylabel is not None:
        grid["ax_margin_y"].set_ylabel(ylabel)

    return grid


def saddle_strength(S, C):
    """
    Parameters
    ----------
    S, C : 2D arrays, square, same shape
        Saddle sums and counts, respectively

    Returns
    -------
    1D array
    Ratios of cumulative corner interaction scores, where the saddle data is
    grouped over the AA+BB corners and AB+BA corners with increasing extent.

    """
    m, n = S.shape
    if m != n:
        raise ValueError("`saddledata` should be square.")

    ratios = np.zeros(n)
    for k in range(1, n):
        intra_sum = np.nansum(S[0:k, 0:k]) + np.nansum(S[n - k : n, n - k : n])
        intra_count = np.nansum(C[0:k, 0:k]) + np.nansum(C[n - k : n, n - k : n])
        intra = intra_sum / intra_count

        inter_sum = np.nansum(S[0:k, n - k : n]) + np.nansum(S[n - k : n, 0:k])
        inter_count = np.nansum(C[0:k, n - k : n]) + np.nansum(C[n - k : n, 0:k])
        inter = inter_sum / inter_count

        ratios[k] = intra / inter
    return ratios



import multiprocess as mp
import click

import cooler

from . import cli
from .. import sample


@cli.command()
@click.argument("in_path", metavar="IN_PATH", type=str, nargs=1)
@click.argument("out_path", metavar="OUT_PATH", type=str, nargs=1)
@click.option(
    "-c",
    "--count",
    help="The target number of contacts in the sample. "
    "The resulting sample size will not match it precisely. "
    "Mutually exclusive with --frac",
    type=int,
    default=None,
    show_default=False,
)
@click.option(
    "-f",
    "--frac",
    help="The target sample size as a fraction of contacts in the original dataset. "
    "Mutually exclusive with --count",
    type=float,
    default=None,
    show_default=False,
)
@click.option(
    "--exact",
    help="If specified, use exact sampling that guarantees the size of the output sample. "
    "Otherwise, binomial sampling will be used and the sample size will be distributed around the target value. ",
    is_flag=True,
)
@click.option(
    "--nproc", "-p",
    help="Number of processes to split the work between."
    "[default: 1, i.e. no process pool]",
    default=1,
    type=int,
)
@click.option(
    "--chunksize",
    help="The number of pixels loaded and processed per step of computation.",
    type=int,
    default=int(1e7),
    show_default=True,
)
def random_sample(in_path, out_path, count, frac, exact, nproc, chunksize):
    """
    Pick a random sample of contacts from a Hi-C map, w/o replacement.

    IN_PATH : Input cooler path or URI.
    
    OUT_PATH : Output cooler path or URI.
    
    Specify the target sample size with either --count or --frac.

    """

    if nproc > 1:
        pool = mp.Pool(nproc)
        map_ = pool.map
    else:
        map_ = map

    try:
        sample.sample_cooler(
            in_path,
            out_path,
            count=count,
            frac=frac,
            exact=exact,
            chunksize=chunksize,
            map_func=map_
        )
    finally:
        if nproc > 1:
            pool.close()



"""
Collection of functions related to dot-calling

"""
from functools import partial, reduce
import multiprocess as mp

from scipy.linalg import toeplitz
from scipy.ndimage import convolve
from scipy.stats import poisson
from scipy.sparse import coo_matrix
import numpy as np
import pandas as pd
from sklearn.cluster import Birch
import cooler

from .lib.numutils import LazyToeplitz, get_kernel

import bioframe

# these are the constants from HiCCUPS, that dictate how initial histogramms
# for lambda-chunking are computed: W1 is the # of logspaced lambda bins,
# and W2 is maximum "allowed" raw number of contacts per Hi-C heatmap bin:
HiCCUPS_W1_MAX_INDX = 40
# there are datasets out there that can exceed this limit.

# we are not using 'W2' as we're building
# the histograms dynamically:
HiCCUPS_W2_MAX_INDX = 10000

# this is to mitigate and parameterize the obs.raw vs count controversy:
observed_count_name = "count"
expected_count_name = "exp.raw"
adjusted_exp_name = lambda kernel_name: "la_exp." + kernel_name + ".value"
nans_inkernel_name = lambda kernel_name: "la_exp." + kernel_name + ".nnans"
bin1_id_name = "bin1_id"
bin2_id_name = "bin2_id"


##################################
# generic service functions:
##################################


def buffer_df_chunks(chunks, size=25000000):
    """
    A buffer that consumes "small" chunks
    and yields larger "chunks" of an
    approximate size "size".

    Following @nvictus:
    https://github.com/mirnylab/cooltools/issues/51#issuecomment-460091252

    """
    buf = []
    n = 0
    for chunk in chunks:
        # accumulate chunks until size
        # is reached:
        n += len(chunk)
        buf.append(chunk)
        if n > size:
            print(
                "large chunk of size {} is ready, made up of {} chunks".format(
                    n, len(buf)
                )
            )
            yield pd.concat(buf, axis=0)
            print("cleaning buffers ...")
            # reset buffer and counter:
            buf = []
            n = 0
    #  yield the remainder of the buffer:
    if len(buf):
        print(
            "LAST chunk of size {} is ready, made up of {} chunks".format(n, len(buf))
        )
        yield pd.concat(buf, axis=0)
        print("last guy shipped ...")


##################################
# dotfinder-specific service functions:
##################################


def recommend_kernel_params(binsize):
    """
    Recommned kernel parameters for the
    standard convolution kernels, 'donut', etc
    same as in Rao et al 2014

    Parameters
    ----------
    binsize : integer
        binsize of the provided cooler

    Returns
    -------
    (w,p) : (integer, integer)
        tuple of the outer and inner kernel sizes
    """
    # kernel parameters depend on the cooler resolution
    # TODO: rename w, p to wid, pix probably, or _w, _p to avoid naming conflicts
    if binsize > 28000:
        # > 30 kb - is probably too much ...
        raise ValueError(
            "Provided cooler has resolution {} bases, "
            "which is too coarse for analysis.".format(binsize)
        )
    elif binsize >= 18000:
        # ~ 20-25 kb:
        w, p = 3, 1
    elif binsize >= 8000:
        # ~ 10 kb
        w, p = 5, 2
    elif binsize >= 4000:
        # ~5 kb
        w, p = 7, 4
    else:
        # < 5 kb - is probably too fine ...
        raise ValueError(
            "Provided cooler has resolution {} bases, "
            "which is too fine for analysis.".format(binsize)
        )
    # return the results:
    return w, p


def annotate_pixels_with_qvalues(
    pixels_df, qvalues, kernels, inplace=False, obs_raw_name=observed_count_name
):
    """
    Add columns with the qvalues to a DataFrame of pixels
    ... detailed but unedited notes ...
    Extract q-values using l-chunks and IntervalIndex.
    we'll do it in an ugly but workign fashion, by simply
    iteration over pairs of obs, la_exp and extracting needed qvals
    one after another
    ...

    Parameters
    ----------
    pixels_df : pandas.DataFrame
        a DataFrame with pixel coordinates that must have at least 2 columns
        named 'bin1_id' and 'bin2_id', where first is pixels's row and the
        second is pixel's column index.
    qvalues : dict of DataFrames
        A dictionary with keys being kernel names and values DataFrames
        storing q-values for each observed count values in each lambda-
        chunk. Colunms are Intervals defined by 'ledges' boundaries.
        Rows corresponding to a range of observed count values.
    kernels : dict
        A dictionary with keys being kernels names and values being ndarrays
        representing those kernels.

    Returns
    -------
    pixels_qvalue_df : pandas.DataFrame
        DataFrame of pixels with additional columns
        storing qvalues corresponding to the observed
        count value of a given pixel, given kernel-type,
        and a lambda-chunk.

    Notes
    -----
    Should be applied to a filtered DF of pixels, otherwise would
    be too resource-hungry.
    """
    if inplace:
        pixels_qvalue_df = pixels_df
    else:
        # let's do it "safe" - using a copy:
        pixels_qvalue_df = pixels_df.copy()
    # attempting to extract q-values using l-chunks and IntervalIndex:
    # we'll do it in an ugly but workign fashion, by simply
    # iteration over pairs of obs, la_exp and extracting needed qvals
    # one after another ...
    for k in kernels:
        pixels_qvalue_df["la_exp." + k + ".qval"] = [
            qvalues[k].loc[o, e]
            for o, e in pixels_df[[obs_raw_name, "la_exp." + k + ".value"]].itertuples(
                index=False
            )
        ]
    # qvalues : dict
    #   A dictionary with keys being kernel names and values pandas.DataFrame-s
    #   storing q-values: each column corresponds to a lambda-chunk,
    #   while rows correspond to observed pixels values.
    return pixels_qvalue_df


def clust_2D_pixels(
    pixels_df,
    threshold_cluster=2,
    bin1_id_name="bin1_id",
    bin2_id_name="bin2_id",
    clust_label_name="c_label",
    clust_size_name="c_size",
    verbose=True,
):
    """
    Group significant pixels by proximity using Birch clustering. We use
    "n_clusters=None", which implies no AgglomerativeClustering, and thus
    simply reporting "blobs" of pixels of radii <="threshold_cluster" along
    with corresponding blob-centroids as well.

    Parameters
    ----------
    pixels_df : pandas.DataFrame
        a DataFrame with pixel coordinates that must have at least 2 columns
        named 'bin1_id' and 'bin2_id', where first is pixels's row and the
        second is pixel's column index.
    threshold_cluster : int
        clustering radius for Birch clustering derived from ~40kb radius of
        clustering and bin size.
    bin1_id_name : str
        Name of the 1st coordinate (row index) in 'pixel_df', by default
        'bin1_id'. 'start1/end1' could be usefull as well.
    bin2_id_name : str
        Name of the 2nd coordinate (column index) in 'pixel_df', by default
        'bin2_id'. 'start2/end2' could be usefull as well.
    clust_label_name : str
        Name of the cluster of pixels label. "c_label" by default.
    clust_size_name : str
        Name of the cluster of pixels size. "c_size" by default.
    verbose : bool
        Print verbose clustering summary report defaults is True.

    Returns
    -------
    peak_tmp : pandas.DataFrame
        DataFrame with the following columns:
        [c+bin1_id_name, c+bin2_id_name, clust_label_name, clust_size_name]
        row/col (bin1/bin2) are coordinates of centroids,
        label and sizes are unique pixel-cluster
        labels and their corresponding sizes.
    """

    # col (bin2) must precede row (bin1):
    pixels = pixels_df[[bin1_id_name, bin2_id_name]].values.astype(np.float)
    # added astype(float) to avoid further issues with clustering, as it
    # turned out start1/start2 genome coordinates could be int32 or int64
    # and int32 is not enough for some operations, i.e., integer overflow.
    pixel_idxs = pixels_df.index

    # perform BIRCH clustering of pixels:
    # "n_clusters=None" implies using BIRCH without AgglomerativeClustering,
    # thus simply reporting "blobs" of pixels of radius "threshold_cluster"
    # along with blob-centroids as well:
    brc = Birch(
        n_clusters=None,
        threshold=threshold_cluster,
        # branching_factor=50, (it's default)
        compute_labels=True,
    )
    brc.fit(pixels)
    # # following is redundant,
    # # as it's done here:
    # # https://github.com/scikit-learn/scikit-learn/blob/a24c8b464d094d2c468a16ea9f8bf8d42d949f84/sklearn/cluster/birch.py#L638
    # clustered_labels = brc.predict(pixels)

    # labels of nearest centroid, assigned to each pixel,
    # BEWARE: labels might not be continuous, i.e.,
    # "np.unique(clustered_labels)" isn't same as "brc.subcluster_labels_", because:
    # https://github.com/scikit-learn/scikit-learn/blob/a24c8b464d094d2c468a16ea9f8bf8d42d949f84/sklearn/cluster/birch.py#L576
    clustered_labels = brc.labels_
    # centroid coordinates ( <= len(clustered_labels)):
    clustered_centroids = brc.subcluster_centers_
    # count unique labels and get their continuous indices
    uniq_labels, inverse_idx, uniq_counts = np.unique(
        clustered_labels, return_inverse=True, return_counts=True
    )
    # cluster sizes taken to match labels:
    cluster_sizes = uniq_counts[inverse_idx]
    # take centroids corresponding to labels (as many as needed):
    centroids_per_pixel = np.take(clustered_centroids, clustered_labels, axis=0)

    if verbose:
        # prepare clustering summary report:
        msg = (
            "Clustering is completed:\n"
            + "{} clusters detected\n".format(uniq_counts.size)
            + "{:.2f}+/-{:.2f} mean size\n".format(
                uniq_counts.mean(), uniq_counts.std()
            )
        )
        print(msg)

    # create output DataFrame
    centroids_n_labels_df = pd.DataFrame(
        centroids_per_pixel,
        index=pixel_idxs,
        columns=["c" + bin1_id_name, "c" + bin2_id_name],
    )
    # add labels per pixel:
    centroids_n_labels_df[clust_label_name] = clustered_labels.astype(np.int)
    # add cluster sizes:
    centroids_n_labels_df[clust_size_name] = cluster_sizes.astype(np.int)

    return centroids_n_labels_df


##################################
# matrix tiling and tiles-generator
##################################


def diagonal_matrix_tiling(start, stop, bandwidth, edge=0, verbose=False):
    """
    Generate a stream of tiling coordinates that guarantee to cover a diagonal
    area of the matrix of size 'bandwidth'. cis-signal only!

    Parameters
    ----------
    start : int
        Starting position of the matrix slice to be tiled (inclusive, bins,
        0-based).
    stop : int
        End position of the matrix slice to be tiled (exclusive, bins, 0-based).
    edge : int
        Small edge around each tile to be included in the yielded coordinates.
    bandwidth : int
        Diagonal tiling is guaranteed to cover a diagonal are of width
        'bandwidth'.

    Yields
    ------
    Pairs of indices for every chunk use those indices [cstart:cstop) to fetch
    chunks from the cooler-object:
    >>> clr.matrix()[cstart:cstop, cstart:cstop]

    Notes
    -----
    Specify a [start,stop) region of a matrix you want to tile diagonally.
    Each slice is characterized by the coordinate of the top-left corner and
    size.

    * * * * * * * * * * *  0-th slice
    *       *           *
    *       *           *
    *   * * * * *       *  i-th slice
    *   *   *   *       *
    * * * * *   *       *
    *   *       *       *
    *   * * * * *       *
    *                   *
    *              ...  *  ...
    *                   *
    *                   *
    * * * * * * * * * * *

    These slices are supposed to cover up a diagonal band of size 'bandwidth'.

    """
    size = stop - start
    tiles = size // bandwidth + bool(size % bandwidth)

    # matrix parameters before chunking:
    if verbose:
        print(
            "matrix of size {}X{} to be split so that\n".format(size, size)
            + "  diagonal region of size {} would be completely\n".format(bandwidth)
            + "  covered by the tiling, additionally keeping\n"
            + "  a small 'edge' of size w={}, to allow for\n".format(edge)
            + "  meaningfull convolution around boundaries.\n"
            + "  Resulting number of tiles is {}".format(tiles - 1)
            + "  Non-edge case size of each tile is {}X{}".format(
                2 * (bandwidth + edge), 2 * (bandwidth + edge)
            )
        )

    # actual number of tiles is tiles-1
    # by doing range(1, tiles) we are making sure we are processing the
    # upper-left chunk only once:
    for t in range(1, tiles):
        # l = max(0,M*t-M)
        # r = min(L,M*t+M)
        lw = max(0, bandwidth * (t - 1) - edge)
        rw = min(size, bandwidth * (t + 1) + edge)
        # don't forget about the 'start' origin:
        yield lw + start, rw + start


def square_matrix_tiling(start, stop, step, edge, square=False, verbose=False):
    """
    Generate a stream of tiling coordinates that guarantee to cover an entire
    matrix. cis-signal only!

    Parameters
    ----------
    start : int
        Starting position of the matrix slice to be tiled (inclusive, bins,
        0-based).
    stop : int
        End position of the matrix slice to be tiled (exclusive, bins, 0-based).
    step : int
        Requested size of the tiles. Boundary tiles may or may not be of
        'step', see 'square'.
    edge : int
        Small edge around each tile to be included in the yielded coordinates.
    square : bool, optional

    Yields
    ------
    Pairs of indices for every chunk use those indices [cstart:cstop) to fetch
    chunks from the cooler-object:
    >>> clr.matrix()[cstart:cstop, cstart:cstop]

    Notes
    -----
    Each slice is characterized by the coordinate
    of the top-left corner and size.

    * * * * * * * * * * * * *
    *       *       *       *
    *       *       *  ...  *
    *       *       *       *
    * * * * * * * * *       *
    *       *               *
    *       *    ...        *
    *       *               *
    * * * * *               *
    *                       *
    *                       *
    *                       *
    * * * * * * * * * * * * *

    Square parameter determines behavior
    of the tiling function, when the
    size of the matrix is not an exact
    multiple of the 'step':

    square = False
    * * * * * * * * * * *
    *       *       *   *
    *       *       *   *
    *       *       *   *
    * * * * * * * * * * *
    *       *       *   *
    *       *       *   *
    *       *       *   *
    * * * * * * * * * * *
    *       *       *   *
    *       *       *   *
    * * * * * * * * * * *
    WARNING: be carefull with extracting
    expected in this case, as it is
    not trivial at all !!!

    square = True
    * * * * * * * * * * *
    *       *   *   *   *
    *       *   *   *   *
    *       *   *   *   *
    * * * * * * * * * * *
    *       *   *   *   *
    *       *   *   *   *
    * * * * * * * * * * *
    * * * * * * * * * * *
    *       *   *   *   *
    *       *   *   *   *
    * * * * * * * * * * *

    """
    size = stop - start
    tiles = size // step + bool(size % step)

    if verbose:
        print(
            "matrix of size {}X{} to be splitted\n".format(size, size)
            + "  into square tiles of size {}.\n".format(step)
            + "  A small 'edge' of size w={} is added, to allow for\n".format(edge)
            + "  meaningfull convolution around boundaries.\n"
            + "  Resulting number of tiles is {}".format(tiles * tiles)
        )

    for tx in range(tiles):
        for ty in range(tiles):

            lwx = max(0, step * tx - edge)
            rwx = min(size, step * (tx + 1) + edge)
            if square and (rwx >= size):
                lwx = size - step - edge

            lwy = max(0, step * ty - edge)
            rwy = min(size, step * (ty + 1) + edge)
            if square and (rwy >= size):
                lwy = size - step - edge

            yield (lwx + start, rwx + start), (lwy + start, rwy + start)


def heatmap_tiles_generator_diag(clr, chroms, pad_size, tile_size, band_to_cover):
    """
    A generator yielding heatmap tiles that are needed to cover the requested
    band_to_cover around diagonal. Each tile is "padded" with pad_size edge to
    allow proper kernel-convolution of pixels close to boundary.

    Parameters
    ----------
    clr : cooler
        Cooler object to use to extract chromosome extents.
    chroms : iterable
        Iterable of chromosomes to process
    pad_size : int
        Size of padding around each tile. Typically the outer size of the
        kernel.
    tile_size : int
        Size of the heatmap tile.
    band_to_cover : int
        Size of the diagonal band to be covered by the generated tiles.
        Typically correspond to the max_loci_separation for called dots.

    Returns
    -------
    tile : tuple
        Generator of tuples of three, which contain
        chromosome name, row index of the tile,
        column index of the tile (chrom, tilei, tilej).

    """

    for chrom in chroms:
        chr_start, chr_stop = clr.extent(chrom)
        for tilei, tilej in square_matrix_tiling(
            chr_start, chr_stop, tile_size, pad_size
        ):
            # check if a given tile intersects with
            # with the diagonal band of interest ...
            diag_from = tilej[0] - tilei[1]
            diag_to = tilej[1] - tilei[0]
            #
            band_from = 0
            band_to = band_to_cover
            # we are using this >2*padding trick to exclude
            # tiles from the lower triangle from calculations ...
            if (min(band_to, diag_to) - max(band_from, diag_from)) > 2 * pad_size:
                yield chrom, tilei, tilej


##################################
# kernel-convolution related:
##################################
def _convolve_and_count_nans(O_bal, E_bal, E_raw, N_bal, kernel):
    """
    Dense versions of a bunch of matrices needed for convolution and
    calculation of number of NaNs in a vicinity of each pixel. And a kernel to
    be provided of course.

    """
    # a matrix filled with the kernel-weighted sums
    # based on a balanced observed matrix:
    KO = convolve(O_bal, kernel, mode="constant", cval=0.0, origin=0)
    # a matrix filled with the kernel-weighted sums
    # based on a balanced expected matrix:
    KE = convolve(E_bal, kernel, mode="constant", cval=0.0, origin=0)
    # get number of NaNs in a vicinity of every
    # pixel (kernel's nonzero footprint)
    # based on the NaN-matrix N_bal.
    # N_bal is shared NaNs between O_bal E_bal,
    # is it redundant ?
    NN = convolve(
        N_bal.astype(np.int),
        # we have to use kernel's
        # nonzero footprint:
        (kernel != 0).astype(np.int),
        mode="constant",
        # there are only NaNs
        # beyond the boundary:
        cval=1,
        origin=0,
    )
    ######################################
    # using cval=0 for actual data and
    # cval=1 for NaNs matrix reduces
    # "boundary issue" to the "number of
    # NaNs"-issue
    # ####################################

    # now finally, E_raw*(KO/KE), as the
    # locally-adjusted expected with raw counts as values:
    Ek_raw = np.multiply(E_raw, np.divide(KO, KE))
    # return locally adjusted expected and number of NaNs
    # in the form of dense matrices:
    return Ek_raw, NN


########################################################################
# this should be a MAIN function to get locally adjusted expected
# Die Hauptfunktion
########################################################################
# ! use parameterized column names
# # observed_count_name = "count"
# # expected_count_name = "exp.raw"
# # adjusted_exp_name = lambda kernel_name: "la_exp."+kernel_name+".value"
# # nans_inkernel_name = lambda kernel_name: "la_exp."+kernel_name+".nnans"
# # bin1_id_name='bin1_id'
# # bin2_id_name='bin2_id'
def get_adjusted_expected_tile_some_nans(
    origin, observed, expected, bal_weights, kernels, balance_factor=None, verbose=False
):
    """
    Get locally adjusted expected for a collection of local-filters (kernels).

    Such locally adjusted expected, 'Ek' for a given kernel,
    can serve as a baseline for deciding whether a given
    pixel is enriched enough to call it a feature (dot-loop,
    flare, etc.) in a downstream analysis.

    For every pixel of interest [i,j], locally adjusted
    expected is a product of a global expected in that
    pixel E_bal[i,j] and an enrichment of local environ-
    ment of the pixel, described with a given kernel:

    ::

                                  KERNEL[i,j](O_bal)
        Ek_bal[i,j] = E_bal[i,j]* ------------------
                                  KERNEL[i,j](E_bal)

    where KERNEL[i,j](X) is a result of convolution
    between the kernel and a slice of matrix X centered
    around (i,j). See link below for details:
    https://en.wikipedia.org/wiki/Kernel_(image_processing)

    Returned values for observed and all expecteds
    are rescaled back to raw-counts, for the sake of
    downstream statistical analysis, which is using
    Poisson test to decide is a given pixel is enriched.
    (comparison between balanced values using Poisson-
    test is intractable):

    ::

                                  KERNEL[i,j](O_bal)
        Ek_raw[i,j] = E_raw[i,j]* ------------------ ,
                                  KERNEL[i,j](E_bal)

    where E_raw[i,j] is:

    ::

              1               1
        -------------- * -------------- * E_bal[i,j]
        bal_weights[i]   bal_weights[j]


    Parameters
    ----------
    origin : (int,int) tuple
        tuple of interegers that specify the
        location of an observed matrix slice.
        Measured in bins, not in nucleotides.
    observed : numpy.ndarray
        square symmetrical dense-matrix
        that contains balanced observed O_bal
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        should we switch to RAW here ?
        it would be easy for the output ....
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    expected : numpy.ndarray
        square symmetrical dense-matrix
        that contains expected, calculated
        based on balanced observed: E_bal.
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        should we switch to its 1D representation here ?
        good for now, but expected might change later ...
        Tanay's expected, expected with modeled TAD's etc ...
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    bal_weights : numpy.ndarray or (numpy.ndarray, numpy.ndarray)
        1D vector used to turn raw observed
        into balanced observed for a slice of
        a matrix with the origin on the diagonal;
        and a tuple/list of a couple of 1D arrays
        in case it is a slice with an arbitrary
        origin.
    kernels : dict of (str, numpy.ndarray)
        dictionary of kernels/masks to perform
        convolution of the heatmap. Kernels
        describe the local environment, and
        used to estimate baseline for finding
        enriched/prominent peaks.
        Peak must be enriched with respect to
        all local environments (all kernels),
        to be considered significant.
        Dictionay keys must contain names for
        each kernel.
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        Beware!: kernels are flipped and
        only then multiplied to matrix by
        scipy.ndimage.convolve
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    balance_factor: float
        Multiplicative Balancing factor:
        balanced matrix multiplied by this factor
        sums up to the total number of reads (taking
        symmetry into account) instead of number of
        bins in matrix. defaults to None and skips
        a lowleft KernelObserved factor calculation.
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
        We need this to test how dynamic donut size
        is affecting peak calling results.
        !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    verbose: bool
        Set to True to print some progress
        messages to stdout.

    Returns
    -------
    peaks_df : pandas.DataFrame
        sparsified DataFrame that stores results of
        locally adjusted calculations for every kernel
        for a given slice of input matrix. Multiple
        instences of such 'peaks_df' can be concatena-
        ted and deduplicated for the downstream analysis.
        Reported columns:
        bin1_id - bin1_id index (row), adjusted to origin
        bin2_id - bin bin2_id index, adjusted to origin
        la_exp - locally adjusted expected (for each kernel)
        la_nan - number of NaNs around (each kernel's footprint)
        exp.raw - global expected, rescaled to raw-counts
        obs.raw - observed values in raw-counts.

    """
    # extract origin coordinate of this tile:
    io, jo = origin
    # let's extract full matrices and ice_vector:
    O_raw = observed  # raw observed, no need to copy, no modifications.
    E_bal = np.copy(expected)
    # 'bal_weights': ndarray or a couple of those ...
    if isinstance(bal_weights, np.ndarray):
        v_bal_i = bal_weights
        v_bal_j = bal_weights
    elif isinstance(bal_weights, (tuple, list)):
        v_bal_i, v_bal_j = bal_weights
    else:
        raise ValueError(
            "'bal_weights' must be an numpy.ndarray"
            "for slices of a matrix with diagonal-origin or"
            "a tuple/list of a couple of numpy.ndarray-s"
            "for a slice of matrix with an arbitrary origin."
        )
    # kernels must be a dict with kernel-names as keys
    # and kernel ndarrays as values.
    if not isinstance(kernels, dict):
        raise ValueError(
            "'kernels' must be a dictionary" "with name-keys and ndarrays-values."
        )

    # balanced observed, from raw-observed
    # by element-wise multiply:
    O_bal = np.multiply(O_raw, np.outer(v_bal_i, v_bal_j))
    # O_bal is separate from O_raw memory-wise.

    # fill lower triangle of O_bal and E_bal with NaNs
    # in order to prevent peak calling from the lower triangle
    # and also to provide fair locally adjusted expected
    # estimation for pixels very close to diagonal, whose
    # "donuts"(kernels) would be crossing the main diagonal.
    # The trickiest thing here would be dealing with the origin: io,jo.
    O_bal[np.tril_indices_from(O_bal, k=(io - jo) - 1)] = np.nan
    E_bal[np.tril_indices_from(E_bal, k=(io - jo) - 1)] = np.nan

    # raw E_bal: element-wise division of E_bal[i,j] and
    # v_bal[i]*v_bal[j]:
    E_raw = np.divide(E_bal, np.outer(v_bal_i, v_bal_j))

    # let's calculate a matrix of common NaNs
    # shared between observed and expected:
    # check if it's redundant ? (is NaNs from O_bal sufficient? )
    N_bal = np.logical_or(np.isnan(O_bal), np.isnan(E_bal))
    # fill in common nan-s with zeroes, preventing
    # NaNs during convolution part of '_convolve_and_count_nans':
    O_bal[N_bal] = 0.0
    E_bal[N_bal] = 0.0
    # think about usinf copyto and where functions later:
    # https://stackoverflow.com/questions/6431973/how-to-copy-data-from-a-numpy-array-to-another
    #
    #
    # we are going to accumulate all the results
    # into a DataFrame, keeping NaNs, and other
    # unfiltered results (even the lower triangle for now):
    i, j = np.indices(O_raw.shape)
    # pack it into DataFrame to accumulate results:
    peaks_df = pd.DataFrame({"bin1_id": i.flatten() + io, "bin2_id": j.flatten() + jo})

    with np.errstate(divide="ignore", invalid="ignore"):
        for kernel_name, kernel in kernels.items():
            ###############################
            # kernel-specific calculations:
            ###############################
            # kernel paramters such as width etc
            # are taken into account implicitly ...
            ########################################
            #####################
            # unroll _convolve_and_count_nans function back
            # for us to test the dynamic donut criteria ...
            #####################
            # Ek_raw, NN = _convolve_and_count_nans(O_bal,
            #                                     E_bal,
            #                                     E_raw,
            #                                     N_bal,
            #                                     kernel)
            # Dense versions of a bunch of matrices needed for convolution and
            # calculation of number of NaNs in a vicinity of each pixel. And a kernel to
            # be provided of course.
            # a matrix filled with the kernel-weighted sums
            # based on a balanced observed matrix:
            KO = convolve(O_bal, kernel, mode="constant", cval=0.0, origin=0)
            # a matrix filled with the kernel-weighted sums
            # based on a balanced expected matrix:
            KE = convolve(E_bal, kernel, mode="constant", cval=0.0, origin=0)
            # get number of NaNs in a vicinity of every
            # pixel (kernel's nonzero footprint)
            # based on the NaN-matrix N_bal.
            # N_bal is shared NaNs between O_bal E_bal,
            NN = convolve(
                N_bal.astype(np.int),
                # we have to use kernel's
                # nonzero footprint:
                (kernel != 0).astype(np.int),
                mode="constant",
                # there are only NaNs
                # beyond the boundary:
                cval=1,
                origin=0,
            )
            ######################################
            # using cval=0 for actual data and
            # cval=1 for NaNs matrix reduces
            # "boundary issue" to the "number of
            # NaNs"-issue
            # ####################################
            # now finally, E_raw*(KO/KE), as the
            # locally-adjusted expected with raw counts as values:
            Ek_raw = np.multiply(E_raw, np.divide(KO, KE))

            # this is the place where we would need to extract
            # some results of convolution and multuplt it by the
            # appropriate factor "cooler._load_attrs(‘bins/weight’)[‘scale’]" ...
            if balance_factor and (kernel_name == "lowleft"):
                peaks_df["factor_balance." + kernel_name + ".KerObs"] = (
                    balance_factor * KO.flatten()
                )
                # KO*balance_factor: to be compared with 16 ...
            if verbose:
                print("Convolution with kernel {} is complete.".format(kernel_name))
            #
            # accumulation into single DataFrame:
            # store locally adjusted expected for each kernel
            # and number of NaNs in the footprint of each kernel
            peaks_df["la_exp." + kernel_name + ".value"] = Ek_raw.flatten()
            peaks_df["la_exp." + kernel_name + ".nnans"] = NN.flatten()
            # do all the filter/logic/masking etc on the complete DataFrame ...
    #####################################
    # downstream stuff is supposed to be
    # aggregated over all kernels ...
    #####################################
    peaks_df["exp.raw"] = E_raw.flatten()
    # obs.raw -> count
    peaks_df["count"] = O_raw.flatten()

    # TO BE REFACTORED/deprecated ...
    # compatibility with legacy API is completely BROKEN
    # post-processing allows us to restore it, see tests,
    # but we pay with the processing speed for it.
    mask_ndx = pd.Series(0, index=peaks_df.index, dtype=np.bool)
    for kernel_name, kernel in kernels.items():
        # accummulating with a vector full of 'False':
        mask_ndx_kernel = ~np.isfinite(peaks_df["la_exp." + kernel_name + ".value"])
        mask_ndx = np.logical_or(mask_ndx_kernel, mask_ndx)

    # returning only pixels from upper triangle of a matrix
    # is likely here to stay:
    upper_band = peaks_df["bin1_id"] < peaks_df["bin2_id"]
    # Consider filling lower triangle of the OBSERVED matrix tile
    # with NaNs, instead of this - we'd need this for a fair
    # consideration of the pixels that are super close to the
    # diagonal and in a case, when the corresponding donut would
    # cross a diagonal line.
    # selecting pixels in relation to diagonal - too far, too
    # close etc, is now shifted to the outside of this function
    # a way to simplify code.

    # return good semi-sparsified DF:
    return peaks_df[~mask_ndx & upper_band].reset_index(drop=True)


##################################
# step-specific dot-calling functions
##################################


def score_tile(
    tile_cij,
    clr,
    cis_exp,
    exp_v_name,
    bal_v_name,
    kernels,
    nans_tolerated,
    band_to_cover,
    balance_factor,
    verbose,
):
    """
    The main working function that given a tile of a heatmap, applies kernels to
    perform convolution to calculate locally-adjusted expected and then
    calculates a p-value for every meaningfull pixel against these l.a. expected
    values.

    Parameters
    ----------
    tile_cij : tuple
        Tuple of 3: chromosome name, tile span row-wise, tile span column-wise:
        (chrom, tile_i, tile_j), where tile_i = (start_i, end_i), and
        tile_j = (start_j, end_j).
    clr : cooler
        Cooler object to use to extract Hi-C heatmap data.
    cis_exp : pandas.DataFrame
        DataFrame with 1 dimensional expected, indexed with 'chrom' and 'diag'.
    exp_v_name : str
        Name of a value column in expected DataFrame
    bal_v_name : str
        Name of a value column with balancing weights in a cooler.bins()
        DataFrame. Typically 'weight'.
    kernels : dict
        A dictionary with keys being kernels names and values being ndarrays
        representing those kernels.
    nans_tolerated : int
        Number of NaNs tolerated in a footprint of every kernel.
    band_to_cover : int
        Results would be stored only for pixels connecting loci closer than
        'band_to_cover'.
    balance_factor : float
        Balancing factor to turn sum of balanced matrix back approximately
        to the number of pairs (used for dynamic-donut criteria mostly).
        use None value to disable dynamic-donut criteria calculation.
    verbose : bool
        Enable verbose output.

    Returns
    -------
    res_df : pandas.DataFrame
        results: annotated pixels with calculated locally adjusted expected
        for every kernels, observed, precalculated pvalues, number of NaNs in
        footprint of every kernels, all of that in a form of an annotated
        pixels DataFrame for eligible pixels of a given tile.

    """
    # unpack tile's coordinates
    chrom, tilei, tilej = tile_cij
    origin = (tilei[0], tilej[0])

    # we have to do it for every tile, because
    # chrom is not known apriori (maybe move outside):
    lazy_exp = LazyToeplitz(cis_exp.loc[chrom][exp_v_name].values)

    # RAW observed matrix slice:
    observed = clr.matrix(balance=False)[slice(*tilei), slice(*tilej)]
    # expected as a rectangular tile :
    expected = lazy_exp[slice(*tilei), slice(*tilej)]
    # slice of balance_weight for row-span and column-span :
    bal_weight_i = clr.bins()[slice(*tilei)][bal_v_name].values
    bal_weight_j = clr.bins()[slice(*tilej)][bal_v_name].values

    # do the convolutions
    result = get_adjusted_expected_tile_some_nans(
        origin=origin,
        observed=observed,
        expected=expected,
        bal_weights=(bal_weight_i, bal_weight_j),
        kernels=kernels,
        balance_factor=balance_factor,
        verbose=verbose,
    )

    # Post-processing filters
    # (1) exclude pixels that connect loci further than 'band_to_cover' apart:
    is_inside_band = result["bin1_id"] > (result["bin2_id"] - band_to_cover)

    # (2) identify pixels that pass number of NaNs compliance test for ALL kernels:
    does_comply_nans = np.all(
        result[["la_exp." + k + ".nnans" for k in kernels]] < nans_tolerated, axis=1
    )
    # so, selecting inside band and nNaNs compliant results:
    # ( drop dropping index maybe ??? ) ...
    res_df = result[is_inside_band & does_comply_nans].reset_index(drop=True)
    # #######################################################################
    # # the following should be rewritten such that we return
    # # opnly bare minimum number of columns per chunk - i.e. annotating is too heavy
    # # to be performed here ...
    # #
    # # I'll do it here - switch off annotation, it'll break "extraction" and other
    # # downstream stuff - but we'll fix it afterwards ...
    # #
    # ########################################################################
    # ########################################################################
    # # consider retiring Poisson testing from here, in case we
    # # stick with l-chunking or opposite - add histogramming business here(!)
    # ########################################################################
    # # do Poisson tests:
    # get_pval = lambda la_exp : 1.0 - poisson.cdf(res_df["obs.raw"], la_exp)
    # for k in kernels:
    #     res_df["la_exp."+k+".pval"] = get_pval( res_df["la_exp."+k+".value"] )
    # # annotate and return
    # return cooler.annotate(res_df.reset_index(drop=True), clr.bins()[:])
    #
    # so return only bin_ids , observed-raw (rename to counts, by Nezar's suggestion)
    # and a bunch of locally adjusted expected estimates - 1 per kernel -
    # that's the bare minimum ...
    #
    # per Nezar's suggestion
    # rename obs.raw -> count
    # this would break A LOT of downstream stuff - but let it be ...
    #
    return res_df[
        ["bin1_id", "bin2_id", "count"] + ["la_exp." + k + ".value" for k in kernels]
    ].astype(dtype={"la_exp." + k + ".value": "float64" for k in kernels})


def histogram_scored_pixels(
    scored_df, kernels, ledges, verbose, obs_raw_name=observed_count_name
):
    """
    An attempt to implement HiCCUPS-like lambda-chunking
    statistical procedure.
    This function aims at building up a histogram of locally
    adjusted expected scores for groups of characterized
    pixels.
    Such histograms are later used to compute FDR thresholds
    for different "classes" of hypothesis (classified by their
    l.a. expected scores).

    Parameters
    ----------
    scored_df : pd.DataFrame
        A table with the scoring information for a group of pixels.
    kernels : dict
        A dictionary with keys being kernels names and values being ndarrays
        representing those kernels.
    ledges : ndarray
        An ndarray with bin lambda-edges for groupping loc. adj. expecteds,
        i.e., classifying statistical hypothesis into lambda-classes.
        Left-most bin (-inf, 1], and right-most one (value,+inf].
    verbose : bool
        Enable verbose output.
    obs_raw_name : str
        Name of the column/field that carry number of counts per pixel,
        i.e. observed raw counts.

    Returns
    -------
    hists : dict of pandas.DataFrame
        A dictionary of pandas.DataFrame with lambda/observed histogram for
        every kernel-type.


    Notes
    -----
    This is just an attempt to implement HiCCUPS-like lambda-chunking.
    So we'll be returning histograms corresponding to the chunks of
    scored pixels.
    Consider modifying/accumulation a globally defined hists object,
    or probably making something like a Feature2D class later on
    where hists would be a class feature and histogramming_step would be
    a method.


    """

    # lambda-chunking implies different 'pval' calculation
    # procedure with a single Poisson expected for all the
    # hypothesis in a same "class", i.e. with the l.a. expecteds
    # from the same histogram bin.

    ########################
    # implementation ideas:
    ########################
    # same observations/hypothesis needs to be classified according
    # to different l.a. expecteds (i.e. for different kernel-types),
    # which can be done with a pandas groupby, something like that:
    # https://stackoverflow.com/questions/21441259
    #
    # after that we could iterate over groups and do np.bincout on
    # the "observed.raw" column (assuming it's of integer type) ...

    hists = {}
    for k in kernels:
        # verbose:
        if verbose:
            print("Building a histogram for kernel-type {}".format(k))
        #  we would need to generate a bunch of these histograms for all of the
        # kernel types:
        # needs to be lambda-binned             : scored_df["la_exp."+k+".value"]
        # needs to be histogrammed in every bin : scored_df["obs.raw"]
        #
        # lambda-bin index for kernel-type "k":
        lbins = pd.cut(scored_df["la_exp." + k + ".value"], ledges)
        # now for each lambda-bin construct a histogramm of "obs.raw":
        obs_hist = {}
        for lbin, grp_df in scored_df.groupby(lbins):
            # check if obs.raw is integer of spome kind (temporary):
            # obs.raw -> count
            assert np.issubdtype(grp_df[obs_raw_name].dtype, np.integer)
            # perform bincounting ...
            obs_hist[lbin] = pd.Series(np.bincount(grp_df[obs_raw_name]))
            # ACHTUNG! assigning directly to empty DF leads to data loss!
            # turn ndarray in Series for ease of handling, i.e.
            # assign a bunch of columns of different sizes to DataFrame.
            #
            # Consider updating global "hists" later on, or implementing a
            # special class for that. Mind that Python multiprocessing
            # "threads" are processes and thus cannot share/modify a shared
            # memory location - deal with it, maybe dask-something ?!
            #
            # turned out that storing W1xW2 for every "thread" requires a ton
            # of memory - that's why different sizes... @nvictus ?
        # store W1x(<=W2) hist for every kernel-type:
        hists[k] = pd.DataFrame(obs_hist).fillna(0).astype(np.integer)
    # return a dict of DataFrames with a bunch of histograms:
    return hists


def determine_thresholds(kernels, ledges, gw_hist, fdr):
    """
    given a 'gw_hist' histogram of observed counts
    for each lambda-chunk for each kernel-type, and
    also given a FDR, calculate q-values for each observed
    count value in each lambda-chunk for each kernel-type.

    Returns
    -------
    threshold_df : dict
      each threshold_df[k] is a Series indexed by la_exp intervals
      (IntervalIndex) and it is all we need to extract "good" pixels from
      each chunk ...
    qvalues : dict
      A dictionary with keys being kernel names and values pandas.DataFrames
      storing q-values: each column corresponds to a lambda-chunk,
      while rows correspond to observed pixels values.


    """
    rcs_hist = {}
    rcs_Poisson = {}
    qvalues = {}
    threshold_df = {}
    for k in kernels:
        # Reverse cumulative histogram for this kernel.
        # First row contains total # of pixels in each lambda-chunk.
        rcs_hist[k] = gw_hist[k].iloc[::-1].cumsum(axis=0).iloc[::-1]

        # Assign a unit Poisson distribution to each lambda-chunk.
        # The expected value is the upper boundary of the lambda-chunk.
        #   poisson.sf = 1 - poisson.cdf, but more precise
        #   poisson.sf(-1,mu) == 1.0, i.e. is equivalent to the
        #   poisson.pmf(gw_hist[k].index, mu)[::-1].cumsum()[::-1]
        rcs_Poisson[k] = pd.DataFrame()
        for mu, column in zip(ledges[1:-1], gw_hist[k].columns):
            renorm_factors = rcs_hist[k].loc[0, column]
            rcs_Poisson[k][column] = renorm_factors * poisson.sf(
                gw_hist[k].index - 1, mu
            )

        # Determine the threshold by checking the value at which 'fdr_diff'
        # first turns positive. Fill NaNs with an "unreachably" high value.
        fdr_diff = fdr * rcs_hist[k] - rcs_Poisson[k]
        very_high_value = len(rcs_hist[k])
        threshold_df[k] = (
            fdr_diff.where(fdr_diff > 0)
            .apply(lambda col: col.first_valid_index())
            .fillna(very_high_value)
            .astype(np.integer)
        )
        # q-values
        # bear in mind some issues with lots of NaNs and Infs after
        # such a brave operation ...
        qvalues[k] = rcs_Poisson[k] / rcs_hist[k]

    return threshold_df, qvalues


def extract_scored_pixels(
    scored_df, kernels, thresholds, ledges, verbose, obs_raw_name=observed_count_name
):
    """
    An attempt to implement HiCCUPS-like lambda-chunking
    statistical procedure.
    Use FDR thresholds for different "classes" of hypothesis
    (classified by their l.a. expected scores), in order to
    extract pixels that stand out.

    Parameters
    ----------
    scored_df : pd.DataFrame
        A table with the scoring information for a group of pixels.
    kernels : dict
        A dictionary with keys being kernel names and values being ndarrays
        representing those kernels.
    thresholds : dict
        A dictionary with keys being kernel names and values pandas.Series
        indexed with Intervals defined by 'ledges' boundaries and storing FDR
        thresholds for observed values.
    ledges : ndarray
        An ndarray with bin lambda-edges for groupping loc. adj. expecteds,
        i.e., classifying statistical hypothesis into lambda-classes.
        Left-most bin (-inf, 1], and right-most one (value,+inf].
    verbose : bool
        Enable verbose output.
    obs_raw_name : str
        Name of the column/field that carry number of counts per pixel,
        i.e. observed raw counts.

    Returns
    -------
    scored_df_slice : pandas.DataFrame
        Filtered DataFrame of pixels extracted applying FDR thresholds.

    Notes
    -----
    This is just an attempt to implement HiCCUPS-like lambda-chunking.

    """
    comply_fdr_list = np.ones(len(scored_df), dtype=np.bool)

    for k in kernels:
        # using special features of IntervalIndex:
        # http://pandas.pydata.org/pandas-docs/stable/advanced.html#intervalindex
        # i.e. IntervalIndex can be .loc-ed with values that would be
        # corresponded with their Intervals (!!!):
        # obs.raw -> count
        comply_fdr_k = (
            scored_df[obs_raw_name].values
            > thresholds[k].loc[scored_df["la_exp." + k + ".value"]].values
        )
        # extracting q-values for all of the pixels takes a lot of time
        # we'll do it externally for filtered_pixels only, in order to save
        # time
        #
        # accumulate comply_fdr_k into comply_fdr_list
        # using np.logical_and:
        comply_fdr_list = np.logical_and(comply_fdr_list, comply_fdr_k)
    # return a slice of 'scored_df' that complies FDR thresholds:
    return scored_df[comply_fdr_list]


##################################
# large CLI-helper functions wrapping smaller step-specific ones:
# basically - the dot-calling steps - ONE PASS DOT-CALLING:
##################################


def scoring_step(
    clr,
    expected,
    expected_name,
    balance_name,
    tiles,
    kernels,
    max_nans_tolerated,
    loci_separation_bins,
    output_path,
    nproc,
    output_mode,
    verbose,
):
    """
    Calculates locally adjusted expected
    for each pixel in a designated area of
    the heatmap and dumps it chunk by chunk
    as an HDF table.
    """
    if verbose:
        print("Preparing to convolve {} tiles:".format(len(tiles)))

    # add very_verbose to supress output from convolution of every tile
    very_verbose = False
    job = partial(
        score_tile,
        clr=clr,
        cis_exp=expected,
        exp_v_name=expected_name,
        bal_v_name=balance_name,
        kernels=kernels,
        nans_tolerated=max_nans_tolerated,
        band_to_cover=loci_separation_bins,
        # do not calculate dynamic-donut criteria
        # for now.
        balance_factor=None,
        verbose=very_verbose,
    )

    if nproc > 1:
        pool = mp.Pool(nproc)
        map_ = pool.imap
        map_kwargs = dict(chunksize=int(np.ceil(len(tiles) / nproc)))
        if verbose:
            print(
                "creating a Pool of {} workers to tackle {} tiles".format(
                    nproc, len(tiles)
                )
            )
    else:
        map_ = map
        if verbose:
            print("fallback to serial implementation.")
        map_kwargs = {}
    try:
        # consider using
        # https://github.com/mirnylab/cooler/blob/9e72ee202b0ac6f9d93fd2444d6f94c524962769/cooler/tools.py#L59
        # here:
        chunks = map_(job, tiles, **map_kwargs)
        ###########################################
        #
        # this is to be rewritten using cooler output
        # as in https://github.com/mirnylab/cooltools/issues/51#issuecomment-458664781
        #
        ###########################################
        if output_mode == "hdf":
            print("pandas hdf output ...")
            append = False
            for chunk in chunks:
                chunk.to_hdf(
                    output_path,
                    key="results",
                    format="table",
                    complevel=9,
                    complib="blosc:snappy",
                    append=append,
                )
                append = True
            # success
            return 0
        # ###########################################
        # #
        # # cooler as an output for the dump:
        # #
        # ###########################################
        elif output_mode == "cooler":
            print("cooler output ...")
            # "create_cooler" API needs more consistency:
            # https://github.com/mirnylab/cooler/issues/149
            #  manually describing "data"-columns (except for bin_ids).
            columns = ["count",] + ["la_exp." + k + ".value" for k in kernels]
            # # dtype them just in case as well:
            # dtypes = {"la_exp."+k+".value":np.float64 for k in kernels}
            # dtypes["count"] = np.int32
            # create_cooler using buffered chunks iterator, to speed up
            # cooler merge, as suggested by @nvictus
            cooler.create_cooler(
                output_path,
                clr.bins()[:],
                # iterator of larger pixel chunks ...
                buffer_df_chunks(chunks),
                columns=columns,
                # dtypes=dtypes,
                # # to be included later, copy from original clr?
                # metadata : dict, optional
                # assembly : str, optional
                ordered=False,  # not ordered for sure
                symmetric_upper=True,  # is it really ? should be
                mode="w",
            )
            # success
            return 0
        # ###########################################
        # #
        # # parquet  - how do I use it with my chunks ?!
        # #
        # ###########################################
        elif output_mode == "parquet":
            print("parquet output ...")
            bioframe.to_parquet(
                chunks,
                output_path,
                # # use defaults first ...
                # row_group_size=None,
                # compression='snappy',
                # use_dictionary=True,
                # version=2.0
            )
            # success
            return 0
        # ###########################################
        # #
        # # local - in memory copy of the dataframe (danger of RAM overuse)
        # #
        # ###########################################
        elif output_mode == "local":
            print("returning local copy of the dataframe ...")
            return pd.concat(chunks)
        else:
            raise ValueError("{} mode is not supported".format(output_mode))
    finally:
        if nproc > 1:
            pool.close()


def histogramming_step(
    scores_input, input_mode, kernels, ledges, output_path=None, nproc=1, verbose=False
):
    """
    This is an implementation of the 1st step of lambda-chunking - histogramming.
    This function expects a BEDPE-style dataframe with the scores to be
    hitogrammed.

    we would need a handle or a file name for the dataframe, list of columns to use

    Parameters
    ----------
    scores_input : file name or handle
        File name or handle of the file with the DataFrame of scores.
        Only parquet files are supported for now.
    input_mode : str
        What type of file is provided as input: parquet, cooler, hdf, etc.
        Only parquet is supported at the moment.
    kernels : dict
        A dictionary with keys being kernel names and values pandas.Series
        indexed with Intervals defined by 'ledges' boundaries and storing FDR
        thresholds for observed values.
        [We should replace it with a simple set of column names  -
        to be histogrammed, or column positions  - make it more generic]
    ledges : ndarray
        An ndarray with bin lambda-edges for groupping loc. adj. expecteds,
        i.e., classifying statistical hypothesis into lambda-classes.
        Left-most bin (-inf, 1], and right-most one (value,+inf].
    output_path : file name or handle
        Where to store an output - i.e. a 2D histogram.
    nproc : int
        Number of workerks to split processing with.
        [reconsider that - as the parquet reading is a bit strange
        in a way it utilizes the cores ...]
    verbose : bool
        Enable verbose output.

    Returns
    -------
    final_hist : pandas.DataFrame
        Filtered DataFrame of pixels extracted applying FDR thresholds.

    Notes
    -----
    This is just an attempt to implement HiCCUPS-like lambda-chunking.
    """
    if verbose:
        print("Preparing to histogram the scores ...")

    # add very_verbose to supress output from convolution of every tile
    very_verbose = False

    if input_mode == "parquet":
        print("parquet input ...")
        # quick attempt - to be updated later
        # ask @nvictus for best practises
        #  wrap it in try statement at least ...
        from pyarrow.parquet import ParquetFile

        # check if scored_input is of str instance
        # or a file handle or whatever ...
        pf = ParquetFile(scores_input)
        # read number of groups
        # still don't know how to control that
        # maybe:
        # "bioframe.to_parquet(... row_group_size = chunksize...)" ?!?
        tiles = range(pf.num_row_groups)
        # gg.read_row_group(i, columns=None,
        #                    nthreads=None, use_threads=True,
        #                    use_pandas_metadata=False)
        extract_df = lambda idx: pf.read_row_group(idx, use_threads=True).to_pandas()
    else:
        raise ValueError("{} mode is not supported".format(input_mode))

    # to hist per scored chunk:
    to_hist = partial(
        histogram_scored_pixels, kernels=kernels, ledges=ledges, verbose=very_verbose
    )

    # composing/piping scoring and histogramming
    # together :
    job = lambda tile: to_hist(extract_df(tile))

    # copy paste from @nvictus modified 'scoring_step':
    if nproc > 1:
        raise NotImplementedError(
            "reading parquet in chunks using multiprocess breaks!"
        )
        # pool = mp.Pool(nproc)
        # map_ = pool.imap
        # map_kwargs = dict(chunksize=int(np.ceil(len(tiles)/nproc)))
        # if verbose:
        #     print("creating a Pool of {} workers to tackle {} tiles".format(
        #             nproc, len(tiles)))
    else:
        map_ = map
        if verbose:
            print("fallback to serial implementation.")
        map_kwargs = {}
    try:
        # consider using
        # https://github.com/mirnylab/cooler/blob/9e72ee202b0ac6f9d93fd2444d6f94c524962769/cooler/tools.py#L59
        # here:
        hchunks = map_(job, tiles, **map_kwargs)
        # hchunks TO BE ACCUMULATED
        # hopefully 'hchunks' would stay in memory
        # until we would get a chance to accumulate them:
    finally:
        if nproc > 1:
            pool.close()
    #
    # now we need to combine/sum all of the histograms
    # for different kernels:
    #
    # assuming we know "kernels"
    # this is very ugly, but ok
    # for the draft lambda-chunking
    # lambda version of lambda-chunking:
    def _sum_hists(hx, hy):
        # perform a DataFrame summation
        # for every value of the dictionary:
        hxy = {}
        for k in kernels:
            hxy[k] = hx[k].add(hy[k], fill_value=0).astype(np.integer)
        # returning the sum:
        return hxy

    # ######################################################
    # this approach is tested and at the very least
    # number of pixels in a dump list matches
    # with the .sum().sum() of the histogram
    # both for 10kb and 5kb,
    # thus we should consider this as a reference
    # implementation, albeit not a very efficient one ...
    # ######################################################
    final_hist = reduce(_sum_hists, hchunks)
    # we have to make sure there is nothing in the
    # top bin, i.e., there are no l.a. expecteds > base^(len(ledges)-1)
    for k in kernels:
        last_la_exp_bin = final_hist[k].columns[-1]
        last_la_exp_vals = final_hist[k].iloc[:, -1]
        # checking the top bin:
        assert (
            last_la_exp_vals.sum() == 0
        ), "There are la_exp.{}.value in {}, please check the histogram".format(
            k, last_la_exp_bin
        )
        # drop that last column/bin (last_edge, +inf]:
        final_hist[k] = final_hist[k].drop(columns=last_la_exp_bin)
        # consider dropping all of the columns that have zero .sum()
    # returning filtered histogram
    return final_hist


def extraction_step(
    scores_input,
    input_mode,
    kernels,
    ledges,
    thresholds,
    nproc=1,
    output_path=None,
    verbose=False,
    bin1_id_name="bin1_id",
    bin2_id_name="bin2_id",
):
    """

    This would be an implementation of the second step of lambda-chunking, i.e.
    extract only "significant" interactions using "count" of each pixel and
    an FDR threshold derived from the BH-FDR, aka histiogramming step.

    we would need a handle or a file name for the dataframe, list of columns to use

    Completion of this step requires pixels counts, which lambda-chunk they belong to,
    and an FDR significant threhsold for a given lambda-chunk.

    Lambda-chunk belonginnes could be provided in different ways:
     - lambda-index
     - the locally-adjusted expected score
         * stored in a dataframe
         * re-computed on the fly (the original-HiCCUPS way)


    Parameters
    ----------
    scores_input : file name or handle
        File name or handle of the file with the DataFrame of scores.
        Only parquet files are supported for now.
    input_mode : str
        What type of file is provided as input: parquet, cooler, hdf, etc.
        Only parquet is supported at the moment.
    kernels : dict
        A dictionary with keys being kernel names and values pandas.Series
        indexed with Intervals defined by 'ledges' boundaries and storing FDR
        thresholds for observed values.
        [We should replace it with a simple set of column names  -
        to be histogrammed, or column positions  - make it more generic]
    ledges : ndarray
        An ndarray with bin lambda-edges for groupping loc. adj. expecteds,
        i.e., classifying statistical hypothesis into lambda-classes.
        Left-most bin (-inf, 1], and right-most one (value,+inf].
    thresholds : dict of DataFrames
        A kernel-indexed dictionary of DataFrames with the thresholds
        that define significant pixels. These DataFrames are indexed
        using Intervals defined by 'ledges' boundaries.
    nproc : int
        Number of workerks to split processing with.
        [reconsider that - as the parquet reading is a bit strange
        in a way it utilizes the cores ...]
    output_path : file name or handle
        File name or handle of the file to store a DataFrame of
        significant pixels. This is typically small enough to fit
        to a csv. So output in a text-format only for now.
    verbose : bool
        Enable verbose output.

    Returns
    -------
    significant_pixels : pandas.DataFrame
        Filtered DataFrame of pixels extracted applying FDR thresholds.

    """
    if verbose:
        print("Preparing to extact significant pixels ...")

    # add very_verbose to supress output from convolution of every tile
    very_verbose = False

    if input_mode == "parquet":
        print("parquet input ...")
        # quick attempt - to be updated later
        # ask @nvictus for best practises
        #  wrap it in try statement at least ...
        from pyarrow.parquet import ParquetFile

        # check if scored_input is of str instance
        # or a file handle or whatever ...
        pf = ParquetFile(scores_input)
        # read number of groups
        # still don't know how to control that
        # maybe:
        # "formats.to_parquet(... row_group_size = chunksize...)" ?!?
        tiles = range(pf.num_row_groups)
        # gg.read_row_group(i, columns=None,
        #                    nthreads=None, use_threads=True,
        #                    use_pandas_metadata=False)
        extract_df = lambda idx: pf.read_row_group(idx, use_threads=True).to_pandas()
    else:
        raise ValueError("{} mode is not supported".format(input_mode))

    # to hist per scored chunk:
    extract_significant = partial(
        extract_scored_pixels,
        kernels=kernels,
        thresholds=thresholds,
        ledges=ledges,
        verbose=very_verbose,
    )

    # composing/piping scoring and histogramming
    # together :
    job = lambda tile: extract_significant(extract_df(tile))

    # copy paste from @nvictus modified 'scoring_step':
    if nproc > 1:
        raise NotImplementedError(
            "reading parquet in chunks using multiprocess breaks!"
        )
    else:
        map_ = map
        if verbose:
            print("fallback to serial implementation.")
        map_kwargs = {}
    try:
        # consider using
        # https://github.com/mirnylab/cooler/blob/9e72ee202b0ac6f9d93fd2444d6f94c524962769/cooler/tools.py#L59
        # here:
        filtered_pix_chunks = map_(job, tiles, **map_kwargs)
        significant_pixels = pd.concat(filtered_pix_chunks, ignore_index=True)
        if output_path is not None:
            significant_pixels.to_csv(
                output_path, sep="\t", header=True, index=False, compression=None
            )
    finally:
        if nproc > 1:
            # pool.close()
            pass
    # there should be no duplicates in the "significant_pixels" DataFrame of pixels:
    significant_pixels_dups = significant_pixels.duplicated()
    assert (
        not significant_pixels_dups.any()
    ), "Duplicated pixels detected during exctraction {}".format(
        significant_pixels[significant_pixels_dups]
    )
    # sort the result just in case and drop its index:
    return significant_pixels.sort_values(by=[bin1_id_name, bin2_id_name]).reset_index(
        drop=True
    )


def clustering_step(
    scores_df,
    expected_chroms,
    dots_clustering_radius,
    verbose,
    obs_raw_name=observed_count_name,
):
    """

    This is a new "clustering" step updated for the pixels processed by lambda-
    chunking multiple hypothesis testing.

    This method assumes that 'scores_df' is a DataFrame with all of the pixels
    that needs to be clustered, thus there is no additional 'comply_fdr' column
    and selection of compliant pixels.

    This step is a clustering-only (using Birch from scikit).

    Parameters
    ----------
    scores_df : pandas.DataFrame
        DataFrame that stores filtered pixels that are ready to be
        clustered, no more 'comply_fdr' column dependency.
    expected_chroms : iterable
        An iterable of chromosomes to be clustered.
    dots_clustering_radius : int
        Birch-clustering threshold.
    verbose : bool
        Enable verbose output.

    Returns
    -------
    centroids : pandas.DataFrame
        Pixels from 'scores_df' annotated with clustering information.

    Notes
    -----
    'dots_clustering_radius' in Birch clustering algorithm corresponds to a
    double the clustering radius in the "greedy"-clustering used in HiCCUPS
    (to be tested).

    """
    # using different bin12_id_names since all
    # pixels are annotated at this point.
    pixel_clust_list = []
    for chrom in expected_chroms:
        # probably generate one big DataFrame with clustering
        # information only and then just merge it with the
        # existing 'scores_df'-DataFrame.
        # should we use groupby instead of 'scores_df['chrom12']==chrom' ?!
        # to be tested ...
        df = scores_df[
            (
                (scores_df["chrom1"].astype(str) == str(chrom))
                & (scores_df["chrom2"].astype(str) == str(chrom))
            )
        ]
        if not len(df):
            continue

        pixel_clust = clust_2D_pixels(
            df,
            threshold_cluster=dots_clustering_radius,
            bin1_id_name="start1",
            bin2_id_name="start2",
            verbose=verbose,
        )
        pixel_clust_list.append(pixel_clust)
    if verbose:
        print("Clustering is over!")
    # concatenate clustering results ...
    # indexing information persists here ...
    pixel_clust_df = pd.concat(pixel_clust_list, ignore_index=False)

    # now merge pixel_clust_df and scores_df DataFrame ...
    # # and merge (index-wise) with the main DataFrame:
    df = pd.merge(
        scores_df, pixel_clust_df, how="left", left_index=True, right_index=True
    )
    #prevents scores_df categorical values (all chroms, including chrM)
    df['chrom1'] = df['chrom1'].astype(str)
    df['chrom2'] = df['chrom2'].astype(str)
    # report only centroids with highest Observed:
    chrom_clust_group = df.groupby(["chrom1", "chrom2", "c_label"])
    centroids = df.loc[chrom_clust_group[obs_raw_name].idxmax()]
    return centroids


# consider switching step names - "extraction" to "FDR-thresholding"
# and "thresholding" to "enrichment" - for final enrichment selection:
def thresholding_step(centroids, obs_raw_name=observed_count_name):
    # (2)
    # filter by FDR, enrichment etc:
    enrichment_factor_1 = 1.5
    enrichment_factor_2 = 1.75
    enrichment_factor_3 = 2.0
    FDR_orphan_threshold = 0.02
    ######################################################################
    # # Temporarily remove orphans filtering, until q-vals are calculated:
    ######################################################################
    enrichment_fdr_comply = (
        (
            centroids[obs_raw_name]
            > enrichment_factor_2 * centroids["la_exp.lowleft.value"]
        )
        & (
            centroids[obs_raw_name]
            > enrichment_factor_2 * centroids["la_exp.donut.value"]
        )
        & (
            centroids[obs_raw_name]
            > enrichment_factor_1 * centroids["la_exp.vertical.value"]
        )
        & (
            centroids[obs_raw_name]
            > enrichment_factor_1 * centroids["la_exp.horizontal.value"]
        )
        & (
            (
                centroids[obs_raw_name]
                > enrichment_factor_3 * centroids["la_exp.lowleft.value"]
            )
            | (
                centroids[obs_raw_name]
                > enrichment_factor_3 * centroids["la_exp.donut.value"]
            )
        )
        & (
            (centroids["c_size"] > 1)
            | (
                (
                    centroids["la_exp.lowleft.qval"]
                    + centroids["la_exp.donut.qval"]
                    + centroids["la_exp.vertical.qval"]
                    + centroids["la_exp.horizontal.qval"]
                )
                <= FDR_orphan_threshold
            )
        )
    )
    # #
    # enrichment_fdr_comply = (
    #     (centroids[obs_raw_name] > enrichment_factor_2 * centroids["la_exp.lowleft.value"]) &
    #     (centroids[obs_raw_name] > enrichment_factor_2 * centroids["la_exp.donut.value"]) &
    #     (centroids[obs_raw_name] > enrichment_factor_1 * centroids["la_exp.vertical.value"]) &
    #     (centroids[obs_raw_name] > enrichment_factor_1 * centroids["la_exp.horizontal.value"]) &
    #     ( (centroids[obs_raw_name] > enrichment_factor_3 * centroids["la_exp.lowleft.value"])
    #         | (centroids[obs_raw_name] > enrichment_factor_3 * centroids["la_exp.donut.value"]) )
    # )
    # use "enrichment_fdr_comply" to filter out
    # non-satisfying pixels:
    out = centroids[enrichment_fdr_comply]

    # # # columns up for grabs, take whatever you need:
    # chrom1
    # start1
    # end1
    # weight1
    # chrom2
    # start2
    # end2
    # weight2
    # la_exp.donut.value
    # la_exp.donut.nnans
    # la_exp.vertical.value
    # la_exp.vertical.nnans
    # la_exp.horizontal.value
    # la_exp.horizontal.nnans
    # la_exp.lowleft.value
    # la_exp.lowleft.nnans
    # la_exp.upright.value
    # la_exp.upright.nnans
    # exp.raw
    # obs.raw
    # la_exp.donut.pval
    # la_exp.vertical.pval
    # la_exp.horizontal.pval
    # la_exp.lowleft.pval
    # la_exp.upright.pval
    # la_exp.donut.qval
    # la_exp.vertical.qval
    # la_exp.horizontal.qval
    # la_exp.lowleft.qval
    # la_exp.upright.qval
    # comply_fdr
    # cstart1
    # cstart2
    # c_label
    # c_size

    # ...
    # to be added to the list of output columns:
    # "factor_balance."+"lowleft"+".KerObs"
    # ...

    # tentaive output columns list:
    columns_for_output = [
        "chrom1",
        "start1",
        "end1",
        "chrom2",
        "start2",
        "end2",
        "cstart1",
        "cstart2",
        "c_label",
        "c_size",
        obs_raw_name,
        # 'exp.raw',
        "la_exp.donut.value",
        "la_exp.vertical.value",
        "la_exp.horizontal.value",
        "la_exp.lowleft.value",
        # "factor_balance.lowleft.KerObs",
        # 'la_exp.upright.value',
        # 'la_exp.upright.qval',
        "la_exp.donut.qval",
        "la_exp.vertical.qval",
        "la_exp.horizontal.qval",
        "la_exp.lowleft.qval",
    ]
    return out[columns_for_output]


##################################
# large CLI-helper functions wrapping smaller step-specific ones:
# basically - the dot-calling steps - ON THE FLY - 2 PASS DOT-CALLING (HiCCUPS-style):
##################################


def scoring_and_histogramming_step(
    clr,
    expected,
    expected_name,
    balance_name,
    tiles,
    kernels,
    ledges,
    max_nans_tolerated,
    loci_separation_bins,
    nproc,
    verbose,
):
    """
    This is a derivative of the 'scoring_step' which is supposed to implement
    the 1st of the lambda-chunking procedure - histogramming.

    Basically we are piping scoring operation together with histogramming into a
    single pipeline of per-chunk operations/transforms.
    """
    if verbose:
        print("Preparing to convolve {} tiles:".format(len(tiles)))

    # add very_verbose to supress output from convolution of every tile
    very_verbose = False

    # to score per tile:
    to_score = partial(
        score_tile,
        clr=clr,
        cis_exp=expected,
        exp_v_name=expected_name,
        bal_v_name=balance_name,
        kernels=kernels,
        nans_tolerated=max_nans_tolerated,
        band_to_cover=loci_separation_bins,
        # do not calculate dynamic-donut criteria
        # for now.
        balance_factor=None,
        verbose=very_verbose,
    )

    # to hist per scored chunk:
    to_hist = partial(
        histogram_scored_pixels, kernels=kernels, ledges=ledges, verbose=very_verbose
    )

    # composing/piping scoring and histogramming
    # together :
    job = lambda tile: to_hist(to_score(tile))

    # copy paste from @nvictus modified 'scoring_step':
    if nproc > 1:
        pool = mp.Pool(nproc)
        map_ = pool.imap
        map_kwargs = dict(chunksize=int(np.ceil(len(tiles) / nproc)))
        if verbose:
            print(
                "creating a Pool of {} workers to tackle {} tiles".format(
                    nproc, len(tiles)
                )
            )
    else:
        map_ = map
        if verbose:
            print("fallback to serial implementation.")
        map_kwargs = {}
    try:
        # consider using
        # https://github.com/mirnylab/cooler/blob/9e72ee202b0ac6f9d93fd2444d6f94c524962769/cooler/tools.py#L59
        # here:
        hchunks = map_(job, tiles, **map_kwargs)
        # hchunks TO BE ACCUMULATED
        # hopefully 'hchunks' would stay in memory
        # until we would get a chance to accumulate them:
    finally:
        if nproc > 1:
            pool.close()
    #
    # now we need to combine/sum all of the histograms
    # for different kernels:
    #
    # assuming we know "kernels"
    # this is very ugly, but ok
    # for the draft lambda-chunking
    # lambda version of lambda-chunking:
    def _sum_hists(hx, hy):
        # perform a DataFrame summation
        # for every value of the dictionary:
        hxy = {}
        for k in kernels:
            hxy[k] = hx[k].add(hy[k], fill_value=0).astype(np.integer)
        # returning the sum:
        return hxy

    # ######################################################
    # this approach is tested and at the very least
    # number of pixels in a dump list matches
    # with the .sum().sum() of the histogram
    # both for 10kb and 5kb,
    # thus we should consider this as a reference
    # implementation, albeit not a very efficient one ...
    # ######################################################
    final_hist = reduce(_sum_hists, hchunks)
    # we have to make sure there is nothing in the
    # top bin, i.e., there are no l.a. expecteds > base^(len(ledges)-1)
    for k in kernels:
        last_la_exp_bin = final_hist[k].columns[-1]
        last_la_exp_vals = final_hist[k].iloc[:, -1]
        # checking the top bin:
        assert (
            last_la_exp_vals.sum() == 0
        ), "There are la_exp.{}.value in {}, please check the histogram".format(
            k, last_la_exp_bin
        )
        # drop that last column/bin (last_edge, +inf]:
        final_hist[k] = final_hist[k].drop(columns=last_la_exp_bin)
        # consider dropping all of the columns that have zero .sum()
    # returning filtered histogram
    return final_hist


def scoring_and_extraction_step(
    clr,
    expected,
    expected_name,
    balance_name,
    tiles,
    kernels,
    ledges,
    thresholds,
    max_nans_tolerated,
    balance_factor,
    loci_separation_bins,
    output_path,
    nproc,
    verbose,
    bin1_id_name="bin1_id",
    bin2_id_name="bin2_id",
):
    """
    This is a derivative of the 'scoring_step' which is supposed to implement
    the 2nd of the lambda-chunking procedure - extracting pixels that are FDR
    compliant.

    Basically we are piping scoring operation together with extraction into a
    single pipeline of per-chunk operations/transforms.

    """
    if verbose:
        print("Preparing to convolve {} tiles:".format(len(tiles)))

    # add very_verbose to supress output from convolution of every tile
    very_verbose = False

    # to score per tile:
    to_score = partial(
        score_tile,
        clr=clr,
        cis_exp=expected,
        exp_v_name=expected_name,
        bal_v_name=balance_name,
        kernels=kernels,
        nans_tolerated=max_nans_tolerated,
        band_to_cover=loci_separation_bins,
        balance_factor=balance_factor,
        verbose=very_verbose,
    )

    # to hist per scored chunk:
    to_extract = partial(
        extract_scored_pixels,
        kernels=kernels,
        thresholds=thresholds,
        ledges=ledges,
        verbose=very_verbose,
    )

    # composing/piping scoring and histogramming
    # together :
    job = lambda tile: to_extract(to_score(tile))

    # copy paste from @nvictus modified 'scoring_step':
    if nproc > 1:
        pool = mp.Pool(nproc)
        map_ = pool.imap
        map_kwargs = dict(chunksize=int(np.ceil(len(tiles) / nproc)))
        if verbose:
            print(
                "creating a Pool of {} workers to tackle {} tiles".format(
                    nproc, len(tiles)
                )
            )
    else:
        map_ = map
        if verbose:
            print("fallback to serial implementation.")
        map_kwargs = {}
    try:
        # consider using
        # https://github.com/mirnylab/cooler/blob/9e72ee202b0ac6f9d93fd2444d6f94c524962769/cooler/tools.py#L59
        # here:
        filtered_pix_chunks = map_(job, tiles, **map_kwargs)
        significant_pixels = pd.concat(filtered_pix_chunks, ignore_index=True)
        if output_path is not None:
            significant_pixels.to_csv(
                output_path, sep="\t", header=True, index=False, compression=None
            )
    finally:
        if nproc > 1:
            pool.close()
    # there should be no duplicates in the "significant_pixels" DataFrame of pixels:
    significant_pixels_dups = significant_pixels.duplicated()
    assert (
        not significant_pixels_dups.any()
    ), "Duplicated pixels detected during exctraction {}".format(
        significant_pixels[significant_pixels_dups]
    )
    # sort the result just in case and drop its index:
    return significant_pixels.sort_values(by=[bin1_id_name, bin2_id_name]).reset_index(
        drop=True
    )


##################################
# OLD functions - to be retired :
##################################
def get_qvals(pvals):
    """
    B&H FDR control procedure: sort a given array of N p-values, determine their
    rank i and then for each p-value compute a corres- ponding q-value, which is
    a minimum FDR at which that given p-value would pass a BH-FDR test.

    Parameters
    ----------
    pvals : array-like
        array of p-values to use for multiple hypothesis testing

    Returns
    -------
    qvals : numpy.ndarray
        array of q-values

    Notes
    -----
    - BH-FDR reminder: given an array of N p-values, sort it in ascending order
      p[1]<p[2]<p[3]<...<p[N], and find a threshold p-value, p[j] for which
      p[j] < FDR*j/N, and p[j+1] is already p[j+1] >= FDR*(j+1)/N. Peaks
      corresponding to p-values p[1]<p[2]<...p[j] are considered significant.

    - Mostly follows the statsmodels implementation:
      http://www.statsmodels.org/dev/_modules/statsmodels/stats/multitest.html

    - Using alpha=0.02 it is possible to achieve called dots similar to
      pre-update status alpha is meant to be a q-value threshold: "qvals <= alpha"

    """
    # NOTE: This is tested and is capable of reproducing
    # `multiple_test_BH` results
    pvals = np.asarray(pvals)
    n_obs = pvals.size
    # determine rank of p-values (1-indexed):
    porder = np.argsort(pvals)
    prank = np.empty_like(porder)
    prank[porder] = np.arange(1, n_obs + 1)
    # q-value = p-value*N_obs/i(rank of a p-value) ...
    qvals = np.true_divide(n_obs * pvals, prank)
    # return the qvals sorted as the initial pvals:
    return qvals


def clustering_step_old(
    scores_file,
    expected_chroms,
    ktypes,
    fdr,
    dots_clustering_radius,
    verbose,
    obs_raw_name=observed_count_name,
):
    """
    This is an old "clustering" step, before lambda-chunking
    was implemented.
    This step actually includes both multiple hypothesis
    testing (its simple genome-wide version of BH-FDR) and
    a clustering step itself (using Birch from scikit).
    This method also assumes 'scores_file' to be an external
    hdf file, and it would try to read the entire file in
    memory.
    """
    res_df = pd.read_hdf(scores_file, "results")

    # do Benjamin-Hochberg FDR multiple hypothesis tests
    # genome-wide:
    for k in ktypes:
        res_df["la_exp." + k + ".qval"] = get_qvals(res_df["la_exp." + k + ".pval"])

    # combine results of all tests:
    res_df["comply_fdr"] = np.all(
        res_df[["la_exp." + k + ".qval" for k in ktypes]] <= fdr, axis=1
    )

    # print a message for timing:
    if verbose:
        print("Genome-wide multiple hypothesis testing is done.")

    # using different bin12_id_names since all
    # pixels are annotated at this point.
    pixel_clust_list = []
    for chrom in expected_chroms:
        # probably generate one big DataFrame with clustering
        # information only and then just merge it with the
        # existing 'res_df'-DataFrame.
        # should we use groupby instead of 'res_df['chrom12']==chrom' ?!
        # to be tested ...
        df = res_df[
            (
                res_df["comply_fdr"]
                & (res_df["chrom1"] == chrom)
                & (res_df["chrom2"] == chrom)
            )
        ]

        pixel_clust = clust_2D_pixels(
            df,
            threshold_cluster=dots_clustering_radius,
            bin1_id_name="start1",
            bin2_id_name="start2",
            verbose=verbose,
        )
        pixel_clust_list.append(pixel_clust)
    if verbose:
        print("Clustering is over!")
    # concatenate clustering results ...
    # indexing information persists here ...
    pixel_clust_df = pd.concat(pixel_clust_list, ignore_index=False)

    # now merge pixel_clust_df and res_df DataFrame ...
    # # and merge (index-wise) with the main DataFrame:
    df = pd.merge(
        res_df[res_df["comply_fdr"]],
        pixel_clust_df,
        how="left",
        left_index=True,
        right_index=True,
    )

    # report only centroids with highest Observed:
    chrom_clust_group = df.groupby(["chrom1", "chrom2", "c_label"])
    centroids = df.loc[chrom_clust_group[obs_raw_name].idxmax()]
    return centroids



import re
import logging
import warnings
import numpy as np
import pandas as pd
import cooler

from .lib._query import CSRSelector
from .lib import peaks, numutils

logging.basicConfig(level=logging.INFO)


def get_n_pixels(bad_bin_mask, window=10, ignore_diags=2):
    """
    Calculate the number of "good" pixels in a diamond at each bin.

    """
    N = len(bad_bin_mask)
    n_pixels = np.zeros(N)
    loc_bad_bin_mask = np.zeros(N, dtype=bool)
    for i_shift in range(0, window):
        for j_shift in range(0, window):
            if i_shift + j_shift < ignore_diags:
                continue

            loc_bad_bin_mask[:] = False
            if i_shift == 0:
                loc_bad_bin_mask |= bad_bin_mask
            else:
                loc_bad_bin_mask[i_shift:] |= bad_bin_mask[:-i_shift]
            if j_shift == 0:
                loc_bad_bin_mask |= bad_bin_mask
            else:
                loc_bad_bin_mask[:-j_shift] |= bad_bin_mask[j_shift:]

            n_pixels[i_shift : (-j_shift if j_shift else None)] += (
                1 - loc_bad_bin_mask[i_shift : (-j_shift if j_shift else None)]
            )
    return n_pixels


def insul_diamond(pixel_query, bins, window=10, ignore_diags=2, norm_by_median=True):
    """
    Calculates the insulation score of a Hi-C interaction matrix.

    Parameters
    ----------
    pixel_query : RangeQuery object <TODO:update description>
        A table of Hi-C interactions. Must follow the Cooler columnar format:
        bin1_id, bin2_id, count, balanced (optional)).
    bins : pandas.DataFrame
        A table of bins, is used to determine the span of the matrix
        and the locations of bad bins.
    window : int
        The width (in bins) of the diamond window to calculate the insulation
        score.
    ignore_diags : int
        If > 0, the interactions at separations < `ignore_diags` are ignored
        when calculating the insulation score. Typically, a few first diagonals
        of the Hi-C map should be ignored due to contamination with Hi-C
        artifacts.
    norm_by_median : bool
        If True, normalize the insulation score by its NaN-median.
    """
    lo_bin_id = bins.index.min()
    hi_bin_id = bins.index.max() + 1
    N = hi_bin_id - lo_bin_id
    sum_counts = np.zeros(N)
    sum_balanced = np.zeros(N)

    n_pixels = get_n_pixels(
        bins.weight.isnull().values, window=window, ignore_diags=ignore_diags
    )

    for chunk_dict in pixel_query.read_chunked():
        chunk = pd.DataFrame(chunk_dict, columns=["bin1_id", "bin2_id", "count"])
        diag_pixels = chunk[chunk.bin2_id - chunk.bin1_id <= (window - 1) * 2]

        diag_pixels = cooler.annotate(diag_pixels, bins[["weight"]])
        diag_pixels["balanced"] = (
            diag_pixels["count"] * diag_pixels["weight1"] * diag_pixels["weight2"]
        )
        valid_pixel_mask = ~diag_pixels["balanced"].isnull().values

        i = diag_pixels.bin1_id.values - lo_bin_id
        j = diag_pixels.bin2_id.values - lo_bin_id

        for i_shift in range(0, window):
            for j_shift in range(0, window):
                if i_shift + j_shift < ignore_diags:
                    continue

                mask = (
                    (i + i_shift == j - j_shift)
                    & (i + i_shift < N)
                    & (j - j_shift >= 0)
                )

                sum_counts += np.bincount(
                    i[mask] + i_shift, diag_pixels["count"].values[mask], minlength=N
                )

                sum_balanced += np.bincount(
                    i[mask & valid_pixel_mask] + i_shift,
                    diag_pixels["balanced"].values[mask & valid_pixel_mask],
                    minlength=N,
                )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        score = sum_balanced / n_pixels

        if norm_by_median:
            score /= np.nanmedian(score)

    return score, n_pixels, sum_balanced, sum_counts


def calculate_insulation_score(
    clr,
    window_bp,
    ignore_diags=None,
    chromosomes=None,
    append_raw_scores=False,
    chunksize=20000000,
    verbose=False,
):
    """Calculate the diamond insulation scores and call insulating boundaries.

    Parameters
    ----------
    clr : cooler.Cooler
        A cooler with balanced Hi-C data.
    window_bp : int or list
        The size of the sliding diamond window used to calculate the insulation
        score. If a list is provided, then a insulation score if done for each
        value of window_bp.
    ignore_diags : int
        The number of diagonals to ignore. If None, equals the number of
        diagonals ignored during IC balancing.
    append_raw_scores : bool
        If True, append columns with raw scores (sum_counts, sum_balanced, n_pixels)
        to the output table.
    verbose : bool
        If True, report real-time progress.

    Returns
    -------
    ins_table : pandas.DataFrame
        A table containing the insulation scores of the genomic bins and
        the insulating boundary strengths.
    """
    if chromosomes is None:
        chromosomes = clr.chromnames

    bin_size = clr.info["bin-size"]
    ignore_diags = (
        ignore_diags
        if ignore_diags is not None
        else clr._load_attrs(clr.root.rstrip("/") + "/bins/weight")["ignore_diags"]
    )

    if isinstance(window_bp, int):
        window_bp = [window_bp]
    window_bp = np.array(window_bp)
    window_bins = window_bp // bin_size

    bad_win_sizes = window_bp % bin_size != 0
    if np.any(bad_win_sizes):
        raise Exception(
            "The window sizes {} has to be a multiple of the bin size {}".format(
                window_bp[bad_win_sizes], bin_size
            )
        )

    # XXX -- Use a delayed query executor
    nbins = len(clr.bins())
    selector = CSRSelector(
        clr.open("r"), shape=(nbins, nbins), field="count", chunksize=chunksize
    )

    ins_chrom_tables = []
    for chrom in chromosomes:
        if verbose:
            logging.info("Processing {}".format(chrom))

        chrom_bins = clr.bins().fetch(chrom)
        ins_chrom = chrom_bins[["chrom", "start", "end"]].copy()
        ins_chrom["is_bad_bin"] = chrom_bins["weight"].isnull()

        # XXX --- Create a delayed selection
        c0, c1 = clr.extent(chrom)
        chrom_query = selector[c0:c1, c0:c1]

        for j, win_bin in enumerate(window_bins):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", RuntimeWarning)
                # XXX -- updated insul_diamond
                ins_track, n_pixels, sum_balanced, sum_counts = insul_diamond(
                    chrom_query, chrom_bins, window=win_bin, ignore_diags=ignore_diags
                )
                ins_track[ins_track == 0] = np.nan
                ins_track = np.log2(ins_track)

            ins_track[~np.isfinite(ins_track)] = np.nan

            ins_chrom["log2_insulation_score_{}".format(window_bp[j])] = ins_track
            ins_chrom["n_valid_pixels_{}".format(window_bp[j])] = n_pixels

            if append_raw_scores:
                ins_chrom["sum_counts_{}".format(window_bp[j])] = sum_counts
                ins_chrom["sum_balanced_{}".format(window_bp[j])] = sum_balanced

        ins_chrom_tables.append(ins_chrom)

    ins_table = pd.concat(ins_chrom_tables)
    return ins_table


def find_boundaries(
    ins_table,
    min_frac_valid_pixels=0.66,
    min_dist_bad_bin=0,
    log2_ins_key="log2_insulation_score_{WINDOW}",
    n_valid_pixels_key="n_valid_pixels_{WINDOW}",
    is_bad_bin_key="is_bad_bin",
):
    """Call insulating boundaries.

    Find all local minima of the log2(insulation score) and calculate their
    chromosome-wide topographic prominence.

    Parameters
    ----------
    ins_table : pandas.DataFrame
        A bin table with columns containing log2(insulation score),
        the number of valid pixels per diamond and (optionally) the mask
        of bad bins.
    min_frac_valid_pixels : float
        The minimal fraction of valid pixels in a diamond to be used in
        boundary picking and prominence calculation.
    min_dist_bad_bin : int
        The minimal allowed distance to a bad bin.
        Ignore bins that have a bad bin closer than this distance.
    log2_ins_key, n_valid_pixels_key : str
        The names of the columns containing log2_insulation_score and
        the number of valid pixels per diamond. When a template
        containing `{WINDOW}` is provided, the calculation is repeated
        for all pairs of columns matching the template.

    Returns
    -------
    ins_table : pandas.DataFrame
        A bin table with appended columns with boundary prominences.
    """

    if min_dist_bad_bin:
        ins_table = pd.concat(
            [
                df.assign(dist_bad_bin=numutils.dist_to_mask(df.is_bad_bin))
                for chrom, df in ins_table.groupby("chrom")
            ]
        )

    if "{WINDOW}" in log2_ins_key:
        windows = set()
        for col in ins_table.columns:
            m = re.match(log2_ins_key.format(WINDOW=r"(\d+)"), col)
            if m:
                windows.add(int(m.groups()[0]))
    else:
        windows = set([None])

    min_valid_pixels = {
        win: ins_table[n_valid_pixels_key.format(WINDOW=win)].max()
        * min_frac_valid_pixels
        for win in windows
    }

    dfs = []
    for chrom, df in ins_table.groupby("chrom"):
        df = df.reset_index(drop=True)
        for win in windows:
            mask = (
                df[n_valid_pixels_key.format(WINDOW=win)].values
                >= min_valid_pixels[win]
            )

            if min_dist_bad_bin:
                mask &= df.dist_bad_bin.values >= min_dist_bad_bin

            ins_track = df[log2_ins_key.format(WINDOW=win)].values[mask]
            poss, proms = peaks.find_peak_prominence(-ins_track)
            ins_prom_track = np.zeros_like(ins_track) * np.nan
            ins_prom_track[poss] = proms

            if win is not None:
                bs_key = "boundary_strength_{win}".format(win=win)
            else:
                bs_key = "boundary_strength"

            df[bs_key] = np.nan
            df.loc[mask, bs_key] = ins_prom_track

        dfs.append(df)
    return pd.concat(dfs)


def _insul_diamond_dense(mat, window=10, ignore_diags=2, norm_by_median=True):
    """
    Calculates the insulation score of a Hi-C interaction matrix.

    Parameters
    ----------
    mat : numpy.array
        A dense square matrix of Hi-C interaction frequencies.
        May contain nans, e.g. in rows/columns excluded from the analysis.

    window : int
        The width of the window to calculate the insulation score.

    ignore_diags : int
        If > 0, the interactions at separations < `ignore_diags` are ignored
        when calculating the insulation score. Typically, a few first diagonals
        of the Hi-C map should be ignored due to contamination with Hi-C
        artifacts.

    norm_by_median : bool
        If True, normalize the insulation score by its NaN-median.

    """
    if ignore_diags:
        mat = mat.copy()
        for i in range(-ignore_diags + 1, ignore_diags):
            numutils.set_diag(mat, np.nan, i)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")

        N = mat.shape[0]
        score = np.nan * np.ones(N)
        for i in range(0, N):
            lo = max(0, i + 1 - window)
            hi = min(i + window, N)
            # nanmean of interactions to reduce the effect of bad bins
            score[i] = np.nanmean(mat[lo : i + 1, i:hi])
        if norm_by_median:
            score /= np.nanmedian(score)
    return score


def _find_insulating_boundaries_dense(
    clr,
    window_bp=100000,
    balance="weight",
    min_dist_bad_bin=2,
    ignore_diags=None,
    chromosomes=None,
):
    """Calculate the diamond insulation scores and call insulating boundaries.

    Parameters
    ----------
    c : cooler.Cooler
        A cooler with balanced Hi-C data.
    window_bp : int
        The size of the sliding diamond window used to calculate the insulation
        score.
    min_dist_bad_bin : int
        The minimal allowed distance to a bad bin. Do not calculate insulation
        scores for bins having a bad bin closer than this distance.
    ignore_diags : int
        The number of diagonals to ignore. If None, equals the number of
        diagonals ignored during IC balancing.

    Returns
    -------
    ins_table : pandas.DataFrame
        A table containing the insulation scores of the genomic bins and
        the insulating boundary strengths.
    """
    if chromosomes is None:
        chromosomes = clr.chromnames

    bin_size = clr.info["bin-size"]
    ignore_diags = (
        ignore_diags
        if ignore_diags is not None
        else clr._load_attrs(clr.root.rstrip("/") + "/bins/weight")["ignore_diags"]
    )
    window_bins = window_bp // bin_size

    if window_bp % bin_size != 0:
        raise Exception(
            "The window size ({}) has to be a multiple of the bin size {}".format(
                window_bp, bin_size
            )
        )

    ins_chrom_tables = []
    for chrom in chromosomes:
        ins_chrom = clr.bins().fetch(chrom)[["chrom", "start", "end"]]
        is_bad_bin = np.isnan(clr.bins().fetch(chrom)["weight"].values)

        m = clr.matrix(balance=balance).fetch(chrom)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore", RuntimeWarning)
            ins_track = _insul_diamond_dense(m, window_bins, ignore_diags)
            ins_track[ins_track == 0] = np.nan
            ins_track = np.log2(ins_track)

        bad_bin_neighbor = np.zeros_like(is_bad_bin)
        for i in range(0, min_dist_bad_bin):
            if i == 0:
                bad_bin_neighbor = bad_bin_neighbor | is_bad_bin
            else:
                bad_bin_neighbor = bad_bin_neighbor | np.r_[[True] * i, is_bad_bin[:-i]]
                bad_bin_neighbor = bad_bin_neighbor | np.r_[is_bad_bin[i:], [True] * i]

        ins_track[bad_bin_neighbor] = np.nan
        ins_chrom["bad_bin_masked"] = bad_bin_neighbor

        ins_track[~np.isfinite(ins_track)] = np.nan

        ins_chrom["log2_insulation_score_{}".format(window_bp)] = ins_track

        poss, proms = peaks.find_peak_prominence(-ins_track)
        ins_prom_track = np.zeros_like(ins_track) * np.nan
        ins_prom_track[poss] = proms
        ins_chrom["boundary_strength_{}".format(window_bp)] = ins_prom_track
        ins_chrom["boundary_strength_{}".format(window_bp)] = ins_prom_track

        ins_chrom_tables.append(ins_chrom)

    ins_table = pd.concat(ins_chrom_tables)
    return ins_table
